#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import str
from builtins import map
from builtins import object
from future import standard_library
standard_library.install_aliases()

import os

from ..exceptions import ServerException
from ..repository.repository import Repository, RepositoryConnection
from ...miniclient import repository as miniserver
import re, urllib.request, urllib.parse, urllib.error
from . import spec
from past.builtins import basestring

READ_ONLY = 'READ_ONLY'

LEGAL_OPTION_TYPES = {READ_ONLY: bool,}

ROOT_CATALOG_NAME = 'root'

class AllegroGraphServer(object):
    """
    The AllegroGraphServer object represents a remote AllegroGraph server on
    the network. It is used to inventory and access the catalogs of that
    server.
    """
    def __init__(self, host=None, port=None, user=None, password=None,
                 cainfo=None, sslcert=None, verifyhost=None, verifypeer=None,
                 protocol=None,
                 proxy=os.environ.get('AGRAPH_PROXY'),
                 **options):
        """
        Define the connection to the AllegroGraph HTTP server.

        Pass either ``user`` and ``password`` for Basic Authentication or
        ``cainfo``, ``sslcert`` values for client X.509 certificate
        authentication.

        :param host: Address of the AllegroGraph server to connect to.
                     This can also include protocol and port
                     (e.g. http://localhost:10035). Can also be specified in
                     the ``AGRAPH_HOST`` environment variable.
                     The default value is ``'localhost'```.
        :type host: string
        :param port: Port on which the server is listening.
                     The default is 10035 if ``protocol`` is ``'http'``
                     and 10036 if it is ``'https'``. Can also be specified in
                     the ``AGRAPH_PORT`` environment variable.
                     If passed explicitly this parameter overrides any value
                     that might have been specified as a part of the ``host`` string.
        :type port: int
        :param protocol: Connection protocol, either ``'http'`` or ``'https'``.
                         The default is ``'http'`` if no SSL parameters are set
                         and `'https'`` otherwise.
                         If passed explicitly this parameter overrides any value
                         that might have been specified as a part of the ``host`` string.
        :type protocol: string
        :param user: Username (when using Basic authentication). Can also be specified in
                     the ``AGRAPH_USER`` environment variable.
        :type user: string
        :param password: Password (when using Basic authentication). Can also be specified in
                         the ``AGRAPH_PASSWORD`` environment variable.
        :type password: string
        :param cainfo: Path to a file or directory containing CA certificates that
                       will be used to validate the server's certificate.
        :type cainfo: string
        :param sslcert: Client certificate path (when using SSL authentication).
        :type sslcert: string
        :param verifyhost: If set to ``0`` it will not be an error if the server's
                           SSL certificate does not match the server's address.
                           The default value is ``2``, meaning that the host name will
                           be validated against the certificate.

                           ..seealso:: https://curl.haxx.se/libcurl/c/CURLOPT_SSL_VERIFYHOST.html
        :type verifyhost: int
        :param verifypeer: If set to ``1`` (the default) the validity of the server's
                           SSL certificate will be checked. Set to ``0`` to disable
                           the validation.

                           ..seealso:: https://curl.haxx.se/libcurl/c/CURLOPT_SSL_VERIFYPEER.html
        :type verifypeer: int
        :param proxy: Proxy specification string. The format is SCHEME://HOST:PORT.
                      Supported schemes are 'http', 'socks4' and 'socks5'.
                      Note that for SOCKS proxies DNS requests are performed by the
                      proxy server.
                      The default value is taken from the AGRAPH_PROXY environment
                      variable.
        :type proxy: string
        :param options: Ignored.
        """
        # Not sure why we accept these, but don't want to change the API at this point.
        del options

        host = host or os.environ.get('AGRAPH_HOST', '127.0.0.1')

        # Check if other arguments were passed as a part of host
        match = re.match(r'^(?:(?P<protocol>https?)://)?'
                         r'(?P<host>[^:]*)(?::(?P<port>[0-9]*))?(?P<tail>.*)$',
                         host)
        if match:
            if protocol is None:
                protocol = match.group('protocol')
            if port is None and match.group('port') is not None:
                port = int(match.group('port'))
            host = match.group('host')
            tail = match.group('tail')
        else:
            tail = ''

        has_https_params = cainfo or sslcert or verifyhost is not None or verifypeer is not None
        if protocol is None:
            protocol = 'https' if has_https_params else 'http'

        if port is None:
            if 'AGRAPH_PORT' in os.environ:
                port = int(os.environ.get('AGRAPH_PORT'))
            else:
                port = 10035 if protocol == 'http' else 10036

        uri = '{protocol}://{host}:{port}{tail}'.format(protocol=protocol, host=host, port=port, tail=tail)

        user = user or os.environ.get('AGRAPH_USER')
        password = password or os.environ.get('AGRAPH_PASSWORD')

        self._client = miniserver.Client(uri, user, password, cainfo, sslcert, verifyhost, verifypeer,
                                         proxy=proxy)

    @property
    def url(self):
        """Return the server's URL."""
        return self._client.url

    @property
    def version(self):
        """Return the server's version as a string."""
        return self._client.getVersion()

    @property
    def versionTuple(self):
        """
        Return the version number as a tuple of integers.

        All non-digit characters from the version string are ignored
        and act only as separators. All trailing zeros are skipped.
        This is compatible with the way in which the Java client
        parses and compares AG version strings.
        """
        components = [int(c) for c in re.findall(r'\d+', self.version)]
        while components and components[-1] == 0:
            components.pop()
        return tuple(components)

    def listCatalogs(self):
        """
        Get the list of catalogs on this server.

        A value of ``None`` will be included in this list to
        represent the root catalog.

        :return: A list of catalog names.
        :rtype: list[string]
        """
        catalogs = self._client.listCatalogs()
        return catalogs

    def openCatalog(self, name=None):
        """
        Open a catalog by name. Pass None to open the root catalog.

        :param name: One of the catalog names from :meth:`listCatalogs`
                     or ``None`` to open the root catalog.
        :return: A catalog object.
        :rtype: Catalog
        """
        # Allow for 'root', None and '' (or anything else that evaluates to
        # False) to mean the root catalog.
        if not name or name == ROOT_CATALOG_NAME:
            name = None

        cats = self.listCatalogs()
        if not name in cats:
            raise ServerException("There is no catalog named '%s' (found %s)"
                % (name, cats))

        return Catalog(name, self, self._client)

    def getInitfile(self):
        """
        Retrieve the contents of the server initialization file.

        The initialization file is a collection of Common Lisp code
        that is executed in every back-end as it is created.
        :return: Init file content.
        :rtype: string
        """
        return self._client.getInitfile()

    def setInitfile(self, content=None, restart=True):
        """
        Replace the current initialization file contents with the
        ``content`` string or remove if None.

        :param content: New init file conent.
        :type content: string
        :param restart: specifies whether any running shared back-ends should
                        be shut down, so that subsequent requests will be handled by
                        back-ends that include the new code. Default is ``True``.
        """
        return self._client.setInitfile(content, restart)

    def openSession(self, spec, autocommit=False, lifetime=None, loadinitfile=False):
        """
        Open a session on a federated, reasoning, or filtered store.

        Use the helper functions in the :mod:`franz.openrdf.sail.spec` module
        to create the spec string.

        :param spec: Session :mod:`specification string <franz.openrdf.sail.spec>`.
        :type spec: string
        :param autocommit: Autocommit mode (default: ``False`` = start a transaction).
        :type autocommit: bool
        :param lifetime: The number of seconds a session can be idle before being terminated.
                         The default value is 300 (5 minutes). The maximum acceptable value
                         is 21600 (6 hours).
        :param loadinitfile: If ``True`` the init file will be loaded into the new session.
                             The default is False.
        :type loadinitfile: bool
        :return: A connection to the new session.
        :rtype: RepositoryConnection
        """
        minirep = self._client.openSession(spec, autocommit=autocommit, lifetime=lifetime, loadinitfile=loadinitfile)
        return RepositoryConnection(Repository(None, None, minirep), is_session=True)

    def listScripts(self):
        """
        Return the list of Sitescripts currently on the server.

        When a user creates a session, they can choose to load one or more
        of these scripts into the session.

        :return: A list of script names.
        :rtype: list[string]
        """
        return self._client.listScripts()

    def addScript(self, module, code):
        """
        Create or replace a sitescript.

        :param module: Script name.
        :type module: string
        :param code: Script content.
        :type code: string
        """
        return self._client.addScript(module, code)

    def deleteScript(self, module):
        """
        Delete a sitescript.
        :param module: Script name.
        :type module: string
        """
        return self._client.deleteScript(module)

    def getScript(self, module):
        """
        Get the body of a sitescript.
        :param module: Script name.
        :type module: string
        :return: Script content.
        :rtype: string
        """
        return self._client.getScript(module)

    def getUserData(self, key):
        """
        Retrieve user data value with given key.

        Users can store arbitrary strings on the server
        using :meth:`setUserData`.

        :param key: Value identifier.
        :type key: str
        :return: Stored string or None if there is no data under specified key.
        :rtype: str|None
        """
        return self._client.getUserData(key)

    def setUserData(self, key, data):
        """
        Set user data with given key.

        This can be used to store arbitrary strings on the server.

        :param key: Value identifier.
        :type key: str
        :param data: Value to be stored.
        :type data: str
        """
        self._client.setUserData(key, data)

    def deleteUserData(self, key):
        """
        Remove user data from the server.

        This removes a value set with :meth:`setUserData`

        :param key: Value identifier.
        :type key: str
        """
        self._client.deleteUserData(key)

    def listUserData(self):
        """
        Get all used keys from the user data store on the server.

        :return: A list of key names.
        :rtype: list[str]
        """
        return self._client.listUserData()

    def openFederated(self, repositories, autocommit=False, lifetime=None, loadinitfile=False):
        """
        Open a session that federates several repositories. The
        repositories argument must be an array containing store
        designators, which can be Repository or RepositoryConnection
        objects, strings (naming a store in the root catalog, or the
        URL of a store), or (storename, catalogname) tuples.

        :param repositories: List of repositories to federate.
        :type repositories: list[string|(string, string)|Repository|RepositoryConnection]
        :param autocommit: Autocommit mode (default: ``False`` = start a transaction).
        :type autocommit: bool
        :param lifetime: The number of seconds a session can be idle before being terminated.
                         The default value is 300 (5 minutes). The maximum acceptable value
                         is 21600 (6 hours).
        :param loadinitfile: If ``True`` the init file will be loaded into the new session.
                             The default is False.
        :type loadinitfile: bool
        :return: A connection to the new session.
        :rtype: RepositoryConnection
        """
        def asRepoString(x):
            if isinstance(x, basestring): return spec.local(x)
            elif isinstance(x, tuple): return spec.local(x[0], x[1])
            elif isinstance(x, Repository): return x.getSpec()
            elif isinstance(x, RepositoryConnection):
                if (x.repository.database_name == None):
                    raise ServerException(str(x) + " is not a RepositoryConnection created by Repository.getConnection() and therefore cannot be federated")
                return x.getSpec()
            else: raise TypeError(str(x) + " is not a valid repository specification.")
        return self.openSession(spec.federate(*list(map(asRepoString, repositories))), autocommit, lifetime, loadinitfile)

    def listUsers(self):
        """
        Returns a list of names of all the users that have been defined.
        """
        return self._client.listUsers()

    def addUser(self, name, password=None):
        """
        Create a new user. Expects a password parameter, which specifies the
        user's password (can be left off when creating the anonymous user).
        """
        assert password is not None or name == 'anonymous'
        self._client.addUser(name, password)

    def deleteUser(self, name):
        """
        Delete a user.
        """
        self._client.deleteUser(name)

    def changeUserPassword(self, name, password):
        """
        Change the password for the given user.
        """
        self._client.changeUserPassword(name, password)

    def listUserAccess(self, name):
        """
        Retrieve the read/write access for a user. This returns a result set,
        each element of which has a read, write, catalog, and repository
        component. The first two are booleans, the latter two strings. For
        permissions granted globally, catalog and repository will have a
        value of "*", for those granted per-catalog, only repository will
        be "*". catalog normally contains the catalog name, but for the rootAs above, but also includes the access granted to roles that this user has.
        catalog "/" is used.
        """
        return self._client.listUserAccess(name)

    def addUserAccess(self, name, read, write, catalog='*', repository='*'):
        """
        Grant read/write access to a user.

        :param read: Whether to grant read access. Defaults to ``False``.
        :type read: bool
        :param write: Whether to grant write access. Defaults to ``False``.
        :type write: bool
        :param catalog: Which catalog to grant the access on. Leave off or pass ``"*"`` to grant
                        access on all catalogs. Use ``"/"`` for the root catalog.
        :type catalog: string
        :param repository: Specifies the repository that access is granted on. Passing ``"*"``,
                           or leaving the parameter off, means all repositories in the
                           given catalog.
        :type repository: string
        """
        self._client.addUserAccess(name, read, write, catalog, repository)

    def deleteUserAccess(self, name, read, write, catalog='*', repository='*'):
        """
        Takes the same parameters as PUT on this URL, but revokes the access instead of granting it.
        """
        self._client.deleteUserAccess(name, read, write, catalog, repository)

    def listUserEffectiveAccess(self, name):
        """
        Like listUserAccess, but also includes the access granted to roles that this user has.
        """
        return self._client.listUserEffectiveAccess(name)

    def listUserPermissions(self, name):
        """
        List the permission flags that have been assigned to a user (any of
        super, eval, session, replication).
        """
        return self._client.listUserPermissions(name)

    def listUserEffectivePermissions(self, name):
        """
        Retrieve the permission flags assigned to the user, or any of its roles.
        """
        return self._client.listUserEffectivePermissions(name)

    def addUserPermission(self, name, _type):
        """
        Assigns the given permission to this user. type should be super, eval,
        replication, or session.
        """
        self._client.addUserPermission(name, _type)

    def deleteUserPermission(self, name, _type):
        """
        Revokes the given permission for this user.
        """
        self._client.deleteUserPermission(name, _type)

    def listRoles(self):
        """
        Returns the names of all defined roles.
        """
        return self._client.listRoles()

    def addRole(self, role):
        """
        Creates a new role.
        """
        self._client.addRole(role)

    def deleteRole(self, role):
        """
        Deletes a role.
        """
        self._client.deleteRole(role)

    def listRolePermissions(self, role):
        """
        Lists the permission flags granted to a role.
        """
        return self._client.listRolePermissions(role)

    def addRolePermission(self, role, _type):
        """
        Grants a role a certain permission. type should be super, eval, or session.
        """
        self._client.addRolePermission(role, _type)

    def deleteRolePermission(self, role, _type):
        """
        Revokes a permission for a role.
        """
        self._client.deleteRolePermission(role, _type)

    def listRoleAccess(self, role):
        """
        Query the access granted to a role. Returns a result in the same
        format as the equivalent service for users.
        """
        return self._client.listRoleAccess(role)

    def addRoleAccess(self, role, read, write, catalog='*', repository='*'):
        """
        Grant read/write access to a role. See here for the parameters
        that are expected.
        """
        return self._client.addRoleAccess(role, read, write, catalog, repository)

    def deleteRoleAccess(self, role, read, write, catalog='*', repository='*'):
        """
        Revoke read/write access for a role. Accepts the same parameters as above.
        """
        self._client.deleteRoleAccess(role, read, write, catalog, repository)

    def listUserRoles(self, name):
        """
        Retrieves a list of role names for this user name.
        """
        return self._client.listUserRoles(name)

    def addUserRole(self, name, role):
        """
        Add a role to a user.
        """
        self._client.addUserRole(name, role)

    def deleteUserRole(self, name, role):
        """
        Remove a role from a user.
        """
        self._client.deleteUserRole(name, role)

    def listUserSecurityFilters(self, name, _type):
        """
        List security filters for user.

        _type is one of "allow", "disallow"
        """
        return self._client.listUserSecurityFilters(name, _type)

    def addUserSecurityFilter(self, name, _type, s=None, p=None, o=None, g=None):
        """
        Add a security filter for the user.

        name - user name
        _type - one of 'allow' or 'disallow'
        s - optional subject
        p - optional predicate
        o - optional predicate
        g - optional graph
        """
        self._client.addUserSecurityFilter(name, _type, s, p, o, g)

    def deleteUserSecurityFilter(self, name, _type, s=None, p=None, o=None, g=None):
        """
        Add a security filter for the user.

        name - user name
        _type - one of 'allow' or 'disallow'
        s - optional subject
        p - optional predicate
        o - optional predicate
        g - optional graph
        """
        self._client.deleteUserSecurityFilter(name, _type, s, p, o, g)

    def listRoleSecurityFilters(self, role, _type):
        """
        List security filters for user.

        _type is one of "allow", "disallow"
        """
        return self._client.listRoleSecurityFilters(role, _type)

    def addRoleSecurityFilter(self, role, _type, s=None, p=None, o=None, g=None):
        """
        Add a security filter for the user.

        role - role name
        _type - one of 'allow' or 'disallow'
        s - optional subject
        p - optional predicate
        o - optional predicate
        g - optional graph
        """
        self._client.addRoleSecurityFilter(role, _type, s, p, o, g)

    def deleteRoleSecurityFilter(self, role, _type, s=None, p=None, o=None, g=None):
        """
        Add a security filter for the user.

        role - role name
        _type - one of 'allow' or 'disallow'
        s - optional subject
        p - optional predicate
        o - optional predicate
        g - optional graph
        """
        self._client.deleteRoleSecurityFilter(role, _type, s, p, o, g)


class Catalog(object):
    """
    Container of multiple repositories (triple stores).
    """
    def __init__(self, name, server, client):
        self.server = server
        self.mini_catalog = client.openCatalogByName(name)
        self._name = name

    def getName(self):
        """Return the catalog name."""
        return self._name

    name = property(getName)

    def listRepositories(self):
        """
        Return a list of names of repositories (triple stores) managed by
        this catalog.
        """
        return self.mini_catalog.listRepositories()

    def deleteRepository(self, name):
        """
        Delete a repository.
        :param name: Repository name.
        :type name: string
        """
        return self.mini_catalog.deleteRepository(name)

    def getRepository(self, name, access_verb):
        """
        Creates or opens a repository.
        
        :param name: Repository name.
        :type name: string
        :param access_verb: Determines mode of operation. Possible values are:

                                - **Repository.RENEW** clears the contents of an existing
                                   repository before opening. If the indicated repository does not
                                   exist, it creates one.
                                -  **Repository.OPEN** opens an existing repository, or throws an
                                   exception if the repository is not found.
                                -  **Repository.ACCESS** opens an existing repository, or creates a
                                   new one if the repository is not found.
                                -  **Repository.CREATE** creates a new repository, or throws an
                                   exception if one by that name already exists.
        :return: A repository object.
        :rtype: Repository
        """
        access_verb = access_verb.upper()
        name = urllib.parse.quote_plus(name)
        exists = name in self.listRepositories();
        if access_verb == Repository.RENEW:
            if exists:
                self.deleteRepository(name)
            return self.createRepository(name)

        if access_verb == Repository.CREATE:
            if exists:
                raise ServerException(
                    "Can't create triple store named '%s' because a store with that name already exists.",
                    name)
            return self.createRepository(name)

        if access_verb == Repository.OPEN:
            if not exists:
                raise ServerException(
                    "Can't open a triple store named '%s' because there is none.", name)

            return Repository(self, name, self.mini_catalog.getRepository(name))

        if access_verb == Repository.ACCESS:
            if not exists:
                return self.createRepository(name)

        return Repository(self, name, self.mini_catalog.getRepository(name))

    def createRepository(self, name, indices=None):
        """
        Creates a new repository with the given name.

        :param name: Repository name.
        :type name: string
        :param indices: If provided, creates a store with the given indices.
                        (e.g. ``["spogi, "gspoi", ...]``)
        :type indices: list[string]
        :return: A repository object.
        :rtype: Repository
        """
        return Repository(self, name, self.mini_catalog.createRepository(name, indices=indices))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from future.builtins import object
from past.builtins import unicode

from future.utils import iteritems

from ..model.value import URI

NS = "http://www.w3.org/2001/XMLSchema#"

class XMLSchema(object):
    """
    A 'static' class containing useful XMLSchema URIs.
    """

    NAMESPACE = NS
    DURATION = URI(namespace=NS, localname="duration")
    DATETIME = URI(namespace=NS, localname="dateTime")
    TIME = URI(namespace=NS, localname="time")
    DATE = URI(namespace=NS, localname="date")
    GYEARMONTH = URI(namespace=NS, localname="gYearMonth")
    GYEAR = URI(namespace=NS, localname="gYear")
    GMONTHDAY = URI(namespace=NS, localname="gMonthDay")
    GDAY = URI(namespace=NS, localname="gDay")
    GMONTH = URI(namespace=NS, localname="gMonth")
    STRING = URI(namespace=NS, localname="string")
    BOOLEAN = URI(namespace=NS, localname="boolean")
    BASE64BINARY = URI(namespace=NS, localname="base64Binary")
    HEXBINARY = URI(namespace=NS, localname="hexBinary")
    FLOAT = URI(namespace=NS, localname="float")
    DECIMAL = URI(namespace=NS, localname="decimal")
    DOUBLE = URI(namespace=NS, localname="double")
    ANYURI = URI(namespace=NS, localname="anyURI")
    QNAME = URI(namespace=NS, localname="QName")
    NOTATION = URI(namespace=NS, localname="NOTATION")
    NORMALIZEDSTRING = URI(namespace=NS, localname="normalizedString")
    TOKEN = URI(namespace=NS, localname="token")
    LANGUAGE = URI(namespace=NS, localname="language")
    NMTOKEN = URI(namespace=NS, localname="NMTOKEN")
    NMTOKENS = URI(namespace=NS, localname="NMTOKENS")
    NAME = URI(namespace=NS, localname="Name")
    NCNAME = URI(namespace=NS, localname="NCName")
    ID = URI(namespace=NS, localname="ID")
    IDREF = URI(namespace=NS, localname="IDREF")
    IDREFS = URI(namespace=NS, localname="IDREFS")
    ENTITY = URI(namespace=NS, localname="ENTITY")
    ENTITIES = URI(namespace=NS, localname="ENTITIES")
    INTEGER = URI(namespace=NS, localname="integer")
    LONG = URI(namespace=NS, localname="long")
    INT = URI(namespace=NS, localname="int")
    SHORT = URI(namespace=NS, localname="short")
    NUMBER = URI(namespace=NS, localname="number")        
    BYTE = URI(namespace=NS, localname="byte")
    NON_POSITIVE_INTEGER = URI(namespace=NS, localname="nonPositiveInteger")
    NEGATIVE_INTEGER = URI(namespace=NS, localname="negativeInteger")
    NON_NEGATIVE_INTEGER = URI(namespace=NS, localname="nonNegativeInteger")
    POSITIVE_INTEGER = URI(namespace=NS, localname="positiveInteger")
    UNSIGNED_LONG = URI(namespace=NS, localname="unsignedLong")
    UNSIGNED_INT = URI(namespace=NS, localname="unsignedInt")
    UNSIGNED_SHORT = URI(namespace=NS, localname="unsignedShort")
    UNSIGNED_BYTE = URI(namespace=NS, localname="unsignedByte")        




################################################################################
# Copyright (c) 2006-2017 Franz Inc.
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################
from franz.openrdf.model.value import URI, BNode
from franz.openrdf.model.literal import Literal
from franz.openrdf.util import strings


def parse_term(string_term):
    """
    Given a string representing a term in ntriples format, return
    a URI, Literal, or BNode.
    """
    if not string_term:
        return string_term

    if string_term.startswith('<'):
        return URI(strings.uriref(string_term))
    elif string_term.startswith('"'):
        return Literal(*strings.literal(string_term))
    elif string_term.startswith('_'):
        return BNode(strings.nodeid(string_term))
    elif string_term.startswith('default-graph'):
        return None

    return Literal(string_term)



"""
Functions and structures related to triple attributes and attribute filters.
"""
from __future__ import unicode_literals
from builtins import object, str
from past.builtins import basestring


from typing import List, Union

from six import with_metaclass, python_2_unicode_compatible


class AttributeDefinition(object):
    """
    Defines basic properties of an attribute, such as its name,
    cardinality and the set of allowed values.

    :ivar name: Unique name of this attribute. Must consist only
        of letters, digits, underscores, dashes and non-ascii characters.
    :ivar allowed_values: Either ``None`` or a list of strings that
        defines the set of legal values for this attribute.
    :ivar ordered: If true then values of this attribute can be compared
        in filters. Values are ordered by their index within
        the ``allowed_values`` list (which is required for ordered attributes).
    :ivar minimum_number: Defines the minimum number of values for this
        attribute that must be associated with each triple.
    :ivar maximum_number: Defines the maximum number of values for this
        attribute that can be associated with each triple.
    """
    def __init__(self, name, allowed_values=None, ordered=False,
                 minimum_number=None, maximum_number=None):
        self.name = name  # type: str
        self.allowed_values = allowed_values  # type: Union[List[str], None]
        self.ordered = ordered  # type: bool
        self.minimum_number = minimum_number  # type: Union[int, None]
        self.maximum_number = maximum_number  # type: Union[int, None]


class ContextAttributeType(type):
    """
    Metaclass for 'context' classes that convert ``Class.something``
    and ``Class['something']`` to ``Class('something')``.
    """
    def __getattr__(cls, name):
        return cls(name)

    def __getitem__(cls, index):
        return cls(index)


@python_2_unicode_compatible
class ContextAttribute(with_metaclass(ContextAttributeType)):
    """
    Superclass for contextual triple attributes used in filters.
    """
    context = ''  # Subclasses must override this

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return '%s.%s' % (self.context, self.name)

    def __lt__(self, other):
        return Lt(self, other)

    def __le__(self, other):
        return Le(self, other)

    def __eq__(self, other):
        return Equal(self, other)

    def __gt__(self, other):
        return Gt(self, other)

    def __ge__(self, other):
        return Ge(self, other)

    def __lshift__(self, other):
        return Subset(self, other)

    def __rshift__(self, other):
        return Superset(self, other)

    def __rlshift__(self, other):
        return Subset(other, self)

    def __rrshift__(self, other):
        return Superset(other, self)


class UserAttribute(ContextAttribute):
    """
    User attribute reference to be used in filters.

    A user attribute name 'foo' can be referenced in three ways:

       - ``UserAttribute('foo')``
       - ``UserAttribute.foo``
       - ``UserAttribute['foo']``

    Filters involving objects of this class can be constructed
    using the following operators:

       - ``<``, ``<=``, ``==``, ``>=`` ,``>`` - note that ``==``
         is translated as ``equals``, not ``attribute=``.
       - ``<<`` and ``>>`` - meaning 'subset' and 'superset'.
    """
    context = 'user'


class TripleAttribute(ContextAttribute):
    """
    Triple attribute reference to be used in filters.

    A triple attribute name 'foo' can be referenced in three ways:

       - ``TripleAttribute('foo')``
       - ``TripleAttribute.foo``
       - ``TripleAttribute['foo']``

    Filters involving objects of this class can be constructed
    using the following operators:

       - ``<``, ``<=``, ``==``, ``>=`` ,``>`` - note that ``==``
         is translated as ``equals``, not ``attribute=``.
       - ``<<`` and ``>>`` - meaning 'subset' and 'superset'.
    """
    context = 'triple'


def quote_and_escape_value(value):
    # type: (str) -> str
    """
    Quote a string so it can be read from Lisp.
    """
    return '"' + value.replace('\\', '\\\\').replace('"', '\\"') + '"'


def attribute_set_to_expr(attribute_set):
    """
    Convert an attribute set (to be used in a filter) to string.
    """
    if isinstance(attribute_set, basestring):
        return quote_and_escape_value(attribute_set)
    elif isinstance(attribute_set, ContextAttribute):
        return "%s.%s" % (attribute_set.context, attribute_set.name)
    else:  # Must be a list of values
        values = " ".join(attribute_set_to_expr(v) for v in attribute_set)
        return "(" + values + ")"


def attribute_filter_to_expr(attribute_filter):
    # type: (AttributeFilter|str) -> str
    """
    Convert an attribute filter to string.
    """
    if isinstance(attribute_filter, AttributeFilter):
        return attribute_filter.to_expr()
    return attribute_filter


@python_2_unicode_compatible
class AttributeFilter(object):
    """
    Superclass of objects that can be used as static attribute filters.

    Note that raw strings are also accepted in this role.

    Objects of this class can be combined using bitwise logical operators.
    """
    def to_expr(self):
        """
        Convert this object to an S-expression and return it as a string.
        :return: S-expression representing this object.
        """
        # type: () -> str
        raise NotImplementedError()

    def __and__(self, other):
        # type: (AttributeFilter|str) -> AttributeFilter
        args = []
        if isinstance(self, And):
            args += self.args
        else:
            args.append(self)
        if isinstance(other, And):
            args += other.args
        else:
            args.append(other)
        return And(*args)

    def __or__(self, other):
        # type: (AttributeFilter|str) -> AttributeFilter
        args = []
        if isinstance(self, Or):
            args += self.args
        else:
            args.append(self)
        if isinstance(other, Or):
            args += other.args
        else:
            args.append(other)
        return Or(*args)

    def __ror__(self, other):
        # type: (str) -> AttributeFilter
        return Or(other, self)

    def __rand__(self, other):
        # type: (str) -> AttributeFilter
        return And(other, self)

    def __invert__(self):
        # type: () -> AttributeFilter
        if isinstance(self, Not):
            return self.args[0]
        return Not(self)

    def __str__(self):
        return self.to_expr()


class SetOp(AttributeFilter):
    """
    An attribute filter created by applying an operator
    to one or more attribute sets.
    """
    op = ''  # Subclasses must override this

    def __init__(self, *args):
        self.args = args

    def to_expr(self):
        sets = " ".join(attribute_set_to_expr(v) for v in self.args)
        return "(%s%s%s)" % (self.op, ' ' if sets else '', sets)


class FilterOp(AttributeFilter):
    """
    An attribute filter created by applying an operator
    to other filters. Note that the arguments can be either
    AttributeFilter instances or strings.
    """
    op = ''  # Subclasses must override this

    def __init__(self, *args):
        # type: (*AttributeFilter|str) -> None
        self.args = args

    def to_expr(self):
        filters = " ".join(attribute_filter_to_expr(v) for v in self.args)
        return "(%s%s%s)" % (self.op, ' ' if filters else '', filters)


class And(FilterOp):
    """
    A conjunction of attribute filters.
    """
    op = 'and'


class Or(FilterOp):
    """
    A disjunction of attribute filters.
    """
    op = 'or'


class Not(FilterOp):
    """
    Inverse attribute filter (matches whatever the argument does not match).
    """
    op = 'not'


class Empty(SetOp):
    """
    A filter that matches if the argument is an empty attribute set.
    """
    op = 'empty'


class Overlap(SetOp):
    """
    A filter that matches if attribute sets given as arguments
    have a common value.
    """
    op = 'overlap'


class Subset(SetOp):
    """
    A filter that matches if the first argument is a subset
    of the second argument (both arguments must be attribute sets).
    """
    op = 'subset'


class Superset(SetOp):
    """
    A filter that matches if the first argument is a supersset
    of the second argument (both arguments must be attribute sets).
    """
    op = 'superset'


class Equal(SetOp):
    """
    A filter that matches if its arguments are identical attribute sets.
    """
    op = 'equal'


class OrderedFilter(SetOp):
    """
    A filter for comparing ordered attributes. Both arguments must be either
    empty or singleton sets representing values of an ordered attribute.
    """
    pass


class Lt(OrderedFilter):
    """
    An ordered filter that matches if the first argument is less
    than the second.
    """
    op = 'attribute-set<'


class Le(OrderedFilter):
    """
    An ordered filter that matches if the first argument is less than
    or equal to the second.
    """
    op = 'attribute-set<='


class Eq(OrderedFilter):
    """
    An ordered filter that matches if the first argument is equal to the second.
    """
    op = 'attribute-set='


class Ge(OrderedFilter):
    """
    An ordered filter that matches if the first argument is greater than
    or equal to the second.
    """
    op = 'attribute-set>='


class Gt(OrderedFilter):
    """
    An ordered filter that matches if the first argument is greater than
    the second.
    """
    op = 'attribute-set>'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103 

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import iso8601
from decimal import Decimal

from future.backports import OrderedDict
from future.utils import python_2_unicode_compatible

from past.builtins import long, unicode

from .value import LITERAL_CMP_KEY, URI, Value
from ..exceptions import IllegalArgumentException
from ..vocabulary.xmlschema import XMLSchema
from ..util import strings

import datetime
from collections import defaultdict


# Needed to temporarily convert times to datetimes to do arithmetic.
# Python time objects do not support adding/subtracting time deltas,
# but datetimes do. We need that operation to normalize times to UTC.
RANDOM_DAY = datetime.date(1984, 8, 26)
RANDOM_DAY_STRING = RANDOM_DAY.isoformat()

def datatype_from_python(value, datatype):
    """
    If 'value' is not a string, convert it into one, and infer its
    datatype, unless 'datatype' is set (i.e., overrides it).
    """
    if isinstance(value, unicode):
        return value, datatype

    if isinstance(value, bytes):
        return unicode(value, "utf-8"), datatype

    ## careful: test for 'bool' must precede test for 'int':
    if isinstance(value, bool):
        return unicode(value).lower(), datatype or XMLSchema.BOOLEAN

    if isinstance(value, int):
        return unicode(value), datatype or XMLSchema.INTEGER

    if isinstance(value, long):
        return unicode(value), datatype or XMLSchema.LONG

    if isinstance(value, float):
        return unicode(value), datatype or XMLSchema.DOUBLE

    if isinstance(value, datetime.datetime):
        # There is an ambiguity for times that occur twice due to
        # DST switches, but that is a problem with Python's standard
        # library and there is nothing we can do about it here.
        if value.utcoffset() is not None:
            value = value.replace(tzinfo=None) - value.utcoffset()
        str_value = value.isoformat() + 'Z'
        return str_value, datatype or XMLSchema.DATETIME

    if isinstance(value, datetime.time):
        if value.utcoffset() is not None:
            dt = datetime.datetime.combine(RANDOM_DAY, value)
            dt -= value.utcoffset()
            # Note: this will strip TZ
            value = dt.time()
        str_value = value.isoformat() + 'Z'
        return str_value, datatype or XMLSchema.TIME

    if isinstance(value, datetime.date):
        return value.isoformat(), datatype or XMLSchema.DATE

    if isinstance(value, Decimal):
        return unicode(value), datatype or XMLSchema.DECIMAL

    return unicode(value), datatype


class Literal(Value):
    """
    Implementation of the Literal class.
    """
    
    def __init__(self, label, datatype=None, language=None):
        Value.__init__(self)
        
        # Uses the properties to set the real values
        self.label, self.datatype = datatype_from_python(label, datatype)
        self.language = language

    def getDatatype(self):
        """The URI representing the datatype for this literal, if there is one""" 
        return self._datatype
    
    def setDatatype(self, datatype):
        """Sets the datatype of the value"""
        if isinstance(datatype, bytes):
            datatype = unicode(datatype, "utf-8")
        if isinstance(datatype, unicode):
            if datatype[0] == '<':
                datatype = datatype[1:-1]
            datatype = URI(datatype)
        elif datatype is not None:
            if not isinstance(datatype, URI):
                datatype = URI(datatype)
            elif datatype.uri is None:
                datatype = None

        self._datatype = datatype # pylint: disable-msg=W0201

    datatype = property(getDatatype, setDatatype)

    def getLanguage(self):
        """The language for this Literal"""
        return self._language
    
    def setLanguage(self, language):
        """Set the language for this Literal"""
        self._language = language.lower() if language else None # pylint: disable-msg=W0201

    language = property(getLanguage, setLanguage)

    def getLabel(self):
        """The label/value for this Literal"""
        return self._label
    
    def setLabel(self, label):
        """Set the label for this Literal"""
        self._label = label # pylint: disable-msg=W0201
    
    def getValue(self):
        """The label/value"""
        return self.label

    label = property(getLabel, setLabel)

    def get_cmp_key(self):
        return LITERAL_CMP_KEY, self.label, self.datatype, self.language
    
    def intValue(self):
        """Convert to int"""
        return int(self._label)
    
    def longValue(self):
        """Convert to long"""
        return long(self._label)
    
    def floatValue(self):
        """Convert to float"""
        return float(self._label)
    
    def booleanValue(self):
        """Convert to bool"""
        return self._label == 'true'
    
    def decimalValue(self):
        """Convert to a decimal"""
        return Decimal(self._label)
    
    def dateValue(self):
        """Convert to date"""
        # iso8601 can parse a date, but it will still be
        # returned as a datetime.datetime object. So we have
        # to extract the date (as a datetime.date object) from it.
        return iso8601.parse_date(self._label).date()

    def datetimeValue(self):
        """Convert to datetime"""
        return iso8601.parse_date(self._label, default_timezone=None)

    def timeValue(self):
        """Convert to time"""
        # iso8601 can't parse a time
        # We trick it by prepending a random date,
        # parsing the result as a datetime and extracting
        # the time from that.
        label = RANDOM_DAY_STRING + 'T' + self._label
        # Return a naive time if there is no timezone.
        # The default is UTC.
        dt = iso8601.parse_date(label, default_timezone=None)
        # Note that time() strips the time zone
        return dt.time().replace(tzinfo=dt.tzinfo)

    ## Returns the {@link XMLGregorianCalendar} value of this literal. A calendar
    ## representation can be given for literals whose label conforms to the
    ## syntax of the following <a href="http://www.w3.org/TR/xmlschema-2/">XML
    ## Schema datatypes</a>: <tt>dateTime</tt>, <tt>time</tt>,
    ## <tt>date</tt>, <tt>gYearMonth</tt>, <tt>gMonthDay</tt>,
    ## <tt>gYear</tt>, <tt>gMonth</tt> or <tt>gDay</tt>.
    def calendarValue(self):
        """calendarValue not useful for Python."""
        raise NotImplementedError("calendarValue")

    def toNTriples(self):
        """
        Return an NTriples representation for this Literal.
        """
        sb = []
        sb.append('"')
        sb.append(strings.encode_ntriple_string(self.getLabel()))
        sb.append('"')
        if self.language:
            sb.append('@')
            sb.append(self.language)
        if self.datatype:
            sb.append("^^")
            sb.append(self.datatype.toNTriples())
        return ''.join(sb)

    def to_json_ld(self):
        """ Converts to an object to be used as a JSON-LD value. """
        if self.language:
            return OrderedDict([
                ('@value', self.label),
                ('@language', self.language)
            ])

        if self.datatype is None or self.datatype == XMLSchema.STRING:
            return self.label

        if self.datatype == XMLSchema.INTEGER:
            return int(self.label)

        return OrderedDict([
            ('@value', self.label),
            ('@type', self.datatype.to_json_ld_key())
        ])

###############################################################################
## Automatic conversion from Literal to Python object
###############################################################################
    def toPython(self):
        """
        Return a Python object representation of this literal.   
        Slightly silly implementation because we implement a conversion table
        and then don't use the conversion functions.     
        """
        return XSDToPython[getattr(self.datatype, "uri", None)](self)


XSDToPython = defaultdict(lambda: Literal.getValue, [
                (XMLSchema.INT.uri, Literal.intValue),
                (XMLSchema.FLOAT.uri, Literal.floatValue), 
                (XMLSchema.DOUBLE.uri, Literal.floatValue), 
                (XMLSchema.LONG.uri, Literal.longValue),
                (XMLSchema.INTEGER.uri, Literal.intValue),
                (XMLSchema.BOOLEAN.uri, Literal.booleanValue),
                (XMLSchema.DATETIME.uri, Literal.datetimeValue),
                (XMLSchema.DATE.uri, Literal.dateValue),
                (XMLSchema.TIME.uri, Literal.timeValue),
                (XMLSchema.DECIMAL.uri, Literal.decimalValue)])


###############################################################################
# Extension to RDF4J API
###############################################################################

class CompoundLiteral(Literal):
    """
    A compound literal represents a range, a geospatial coordinate,
    or other useful compound structure.
    TODO: FIGURE OUT SYNTAX FOR OTHER TYPES. INSURE THAT
    THE SYNTAX FOR A RANGE DOESN'T CONFLICT/OVERLAP
    """
    RANGE_LITERAL = 'rangeLiteral'
    def __init__(self, choice, lowerBound=None, upperBound=None):
        self.choice = choice
        if choice == CompoundLiteral.RANGE_LITERAL:
            self.lowerBound = lowerBound # should be a LiteralImpl
            self.upperBound = upperBound # should be a LiteralImpl
        ## other compound types go here.
        else:
            raise IllegalArgumentException("Can't interpret the choice '%s' of a compound literal." % choice)
    
    def isRangeLiteral(self):
        return self.choice == CompoundLiteral.RANGE_LITERAL
    
    def getLowerBound(self):
        return self.lowerBound
    
    def getUpperBound(self):
        return self.upperBound
    
class RangeLiteral(CompoundLiteral):
    """
    A range represents an interval between to scalar values.
    """
    def __init__(self, lowerBound=None, upperBound=None):
        self.lowerBound = lowerBound # should be a LiteralImpl
        self.upperBound = upperBound # should be a LiteralImpl
    
    def getLowerBound(self):
        return self.lowerBound
    
    def getUpperBound(self):
        return self.upperBound

@python_2_unicode_compatible
class GeoCoordinate(CompoundLiteral):
    """
    Define either a cartesian coordinate or a spherical coordinate.  For the
    latter, nit can be 'km', 'mile', 'radian', or 'degree'
    """
    def __init__(self, x, y, unit=None, geoType=None):
        self.xcoor = x
        self.ycoor = y
        self.unit = unit
        self.geoType = geoType
    
    def __str__(self):
        return "|COOR|(%i, %i)" % (self.xcoor, self.ycoor)
    

class GeoSpatialRegion(CompoundLiteral):
    pass


@python_2_unicode_compatible
class GeoBox(GeoSpatialRegion):
    def __init__(self, xMin, xMax, yMin, yMax, unit=None, geoType=None):
        self.xMin = xMin
        self.xMax = xMax
        self.yMin = yMin
        self.yMax = yMax
        self.unit = unit
        self.geoType = geoType
    
    def __str__(self): return "|Box|%s,%s %s,%s" % (self.xMin, self.xMax, self.yMin, self.yMax)


@python_2_unicode_compatible
class GeoCircle(GeoSpatialRegion):
    def __init__(self, x, y, radius, unit=None, geoType=None):
        self.x = x
        self.y = y
        self.radius = radius
        self.unit = unit
        self.geoType=geoType
        
    def __str__(self): return "|Circle|%i,%i, radius=%i" % (self.x, self.y, self.radius)


@python_2_unicode_compatible
class GeoPolygon(GeoSpatialRegion):
    def __init__(self, vertices, uri=None, geoType=None):
        self.vertices = vertices
        self.geoType = geoType
        self.uri = uri
        self.resource = None
        self.miniPolygon = None
        
    def getVertices(self): return self.vertices
    
    def getResource(self): return self.resource
    
    def __str__(self): return "|Polygon|%s" % self.vertices



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from past.builtins import basestring
from builtins import object
from future import standard_library
standard_library.install_aliases()

from ..exceptions import IllegalArgumentException
from ..model import URI, ValueFactory
from .repositoryconnection import RepositoryConnection
from ..vocabulary.xmlschema import XMLSchema

import re


class Repository(object):
    """
    A repository contains RDF data that can be queried and updated. Access
    to the repository can be acquired by opening a connection to it. This
    connection can then be used to query and/or update the contents of the
    repository.

    Please note that a repository needs to be initialized before it can be
    used and that it should be shut down before it is discarded/garbage
    collected. Forgetting the latter can result in loss of data (depending
    on the Repository implementation)!
    """
    # Modes for Catalog.getRepository()
    RENEW = 'RENEW'
    ACCESS = 'ACCESS'
    OPEN = 'OPEN'
    CREATE = 'CREATE'
    REPLACE = 'REPLACE'

    def __init__(self, catalog, database_name, repository):
        """
        Invoke through :meth:`~franz.openrdf.sail.allegrographserver.Catalog.getRepository`.
        """
        self.mini_repository = repository
        self.database_name = database_name
        self.catalog = catalog
        # system state fields:
        self.value_factory = None
         
    def getDatabaseName(self):
        """
        Return the name of the database (remote triple store) that this repository is
        interfacing with.
        """ 
        return self.database_name

    def getSpec(self):
        """
        Return a session spec string for this repository.

        See :meth:`~franz.openrdf.sail.allegrographserver.AllegroGraphServer.openSession`.
        :return: A session spec.
        :rtype: string
        """
        mini = self.mini_repository
        urlstart = re.match("^https?://", mini.url).group(0)
        url = "<%s%s:%s@%s>" % (urlstart, mini.user, mini.password,
            mini.url[len(urlstart):])

        return url
        
    def initialize(self):
        """
        Initializes this repository. A repository must be initialized before
        it can be used.

        It is recommended to take advantage of the fact that repositories
        are context managers and use the ``with`` statement to ensure that
        :meth:`initialize` and :meth:`shutDown` are called:

        .. code:: python

           with catalog.getRepository() as repo:
               # No need to call initialize or shutDown inside.
               ...

        :return: ``self`` (to allow call chaining).
        """
        # We've only kept this method for RDF4J compatibility, we do not actually
        # need to do anything here.
        return self

    def registerDatatypeMapping(self, predicate=None, datatype=None, nativeType=None):
        """
        Register an inlined datatype.

        This allows some literals to be stored in an optimized form
        on the server.

        .. seealso::

           http://franz.com/agraph/support/documentation/current/lisp-reference.html#ref-type-mapping
              More detailed discussion of type mappings in the Lisp API documentation.

        You must supply ``nativeType`` and either ``predicate`` or ``datatype``.

        If ``predicate`` is supplied, then object arguments to triples with that
        predicate will use an inlined encoding of type `nativeType` in their internal
        representation on the server.

        If ``datatype`` is supplied, then typed literal objects with a datatype matching
        ``datatype`` will use an inlined encoding of type `nativeType`.

        Duplicated in the :class:`.RepositoryConnection` class for Python user convenience.

        :param predicate: The URI of a predicate used in the triple store.
        :param datatype: May be one of: ``XMLSchema.INT``, ``XMLSchema.LONG``,
                         ``XMLSchema.FLOAT``, ``XMLSchema.DATE`` and ``XMLSchema.DATETIME``.
        :param nativeType: may be ``int``, ``datetime``, or ``float``.
        :type nativeType: string|type
        """
        predicate = predicate.getURI() if isinstance(predicate, URI) else predicate
        datatype = datatype.getURI() if isinstance(datatype, URI) else datatype

        if nativeType is not None and not isinstance(nativeType, basestring):
            nativeType=nativeType.__name__

        def translate_inlined_type(the_type):
            if the_type == 'int':
                return XMLSchema.LONG.toNTriples()
            if the_type == 'datetime':
                return XMLSchema.DATETIME.toNTriples()
            if the_type == 'time':
                return XMLSchema.TIME.toNTriples()
            if the_type == 'date':
                return XMLSchema.DATE.toNTriples()
            if the_type == "float":
                return XMLSchema.DOUBLE.toNTriples()
            if the_type == "bool":
                return XMLSchema.BOOLEAN.toNTriples()
            raise IllegalArgumentException("Unknown inlined type '%s'\n.  Legal types are "\
                    "int, float, bool, datetime, time, and date." % the_type)
            
        if predicate:
            if not nativeType:
                raise IllegalArgumentException("Missing 'nativeType' parameter in call to 'registerDatatypeMapping'")
            xsdType = translate_inlined_type(nativeType)
            self.mini_repository.addMappedPredicate("<%s>" % predicate, xsdType)            
        elif datatype: 
            xsdType = translate_inlined_type(nativeType or datatype)
            self.mini_repository.addMappedType("<%s>" % datatype, xsdType)
        
    def shutDown(self):
        """
        Shuts the store down, releasing any resources that it keeps hold of.
        Once shut down, the store can no longer be used.

        It is recommended to take advantage of the fact that repositories
        are context managers and use the ``with`` statement to ensure that
        :meth:`initialize` and :meth:`shutDown` are called:

        .. code:: python

           with catalog.getRepository() as repo:
               # No need to call initialize or shutDown inside.
               ...
        """
        self.mini_repository = None

    def isWritable(self):
        """
        Checks whether this store is writable, i.e. if the data contained in
        this store can be changed. The writability of the store is
        determined by the writability of the Sail that this store operates
        on.
        """
        # TODO maybe remove this, it's nonsense in 4.0.
        return True

    def getConnection(self):
        """
        Opens a connection to this store that can be used for querying and
        updating the contents of the store. Created connections need to be
        closed to make sure that any resources they keep hold of are released.
        The best way to ensure this is to use a ``with`` statement:

        .. code:: python

           with repo.getConnection() as conn:
               ...

        :return: A :class:`RepositoryConnection` object.
        :rtype: RepositoryConnection
        """
        return RepositoryConnection(self)

    def getValueFactory(self):
        """
        Return a ValueFactory for this store.

        This is present for RDF4J compatibility, but in the Python API all ValueFactory
        functionality has been duplicated or subsumed in the :class:`.RepositoryConnection` class.
        It isn't necessary to manipulate the :class:`.ValueFactory` class at all.

        :return: A ValueFactory instance.
        :rtype: ValueFactory
        """
        if not self.value_factory:
            self.value_factory = ValueFactory(self)
        return self.value_factory

    def _set_bulk_mode(self, on):
        self.mini_repository.setBulkMode(on)

    def _get_bulk_mode(self):
        return self.mini_repository.getBulkMode()

    bulk_mode = property(
        _get_bulk_mode, _set_bulk_mode,
        doc="""Turn BulkMode on with True or off with False.

               In bulk mode, all modifications to the triple-store are made without
               writing to the transaction log. There is overhead to switching
               in and out of bulk-mode, and it is a global repository state, so all
               clients are affected.""")

    def __enter__(self):
        self.initialize()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del exc_type, exc_val, exc_tb
        self.shutDown()



################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################
from __future__ import unicode_literals
from builtins import range
from builtins import object

from future.utils import python_2_unicode_compatible
from past.builtins import unicode

from ..model import Statement, parse_term
from .repositoryresult import RepositoryResult

from collections import namedtuple


try:
    import franz.openrdf.query.pandas_support as pandas
    has_pandas = True
except ImportError:
    has_pandas = False


class QueryResult(object):
    """
    Super type of all query result types (TupleQueryResult, GraphQueryResult, etc.
    Evaluates as a Python iterator
    """
    def close(self):
        """
        Release resources used by this query result.
        """
        pass

    def __next__(self):
        """
        Return the next result item if there is one.

        :return: The next item.
        :raises StopIteration: If there are no more items to return.
        """
        raise NotImplemented('__next__')

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del exc_type, exc_val, exc_tb
        self.close()

    def __del__(self):
        self.close()


#############################################################################
##
#############################################################################

class GraphQueryResult(RepositoryResult, QueryResult):
    """
    A graph query result is an iterator over the Statements.
    """
    def __init__(self, string_tuples):
        QueryResult.__init__(self)
        RepositoryResult.__init__(self, string_tuples)


class TupleQueryResult(QueryResult):
    """
    A representation of a variable-binding query result as a sequence of
    BindingSet objects. Each query result consists of zero or more
    solutions, each of which represents a single query solution as a set of
    bindings. Note: take care to always close a TupleQueryResult after use to
    free any resources it keeps hold of.
    """
    def __init__(self, variable_names, string_tuples, metadata=None):
        QueryResult.__init__(self)
        if not isinstance(variable_names, list):
            variable_names = [variable_names]
            string_tuples = [string_tuples]
        self.variable_names = variable_names
        self.string_tuples = string_tuples
        self.cursor = 0
        self.tuple_count = len(string_tuples)
        self.binding_set = ListBindingSet(self.variable_names)
        self.metadata = metadata

    def __iter__(self):
        return self

    def __next__(self):
        """
        Return the next result tuple if there is one.

        :return: The next tuple.
        :rtype: ListBindingSet
        :raises StopIteration: If there are no more items to return.
        """
        if self.cursor >= self.tuple_count:
            raise StopIteration()

        bset = self.binding_set
        bset._reset(self.string_tuples[self.cursor])
        self.cursor += 1
        return bset

    def close(self):
        pass

    def getBindingNames(self):
        """
        Get the names of the bindings, in order of projection.

        :return: A list of names.
        :rtype: list[string]
        """
        return self.variable_names

    def getMetadata(self):
        """
        Get a nested dictionary containing query result metadata.

        :return: A dictionary
        :rtype: dict
        """
        return self.metadata

    def __len__(self):
        return self.tuple_count

    def rowCount(self):
        return len(self)

    def toPandas(self):
        if not has_pandas:
            raise Exception('Pandas not installed.')
        return pandas.rows_to_pandas(self, self.variable_names)


@python_2_unicode_compatible
class ListBindingSet(object):
    """
    A BindingSet is a set of named value bindings, which is used to
    represent a single query solution. Values are indexed by name of
    the binding which typically corresponds to the names of the
    variables used in the projection of the original query.

    ListBindingSet emulates an RDF4J BindingSet, a Python dictionary
    and a list simultaneously.  The internal datastructure is a pair
    of lists.
    """
    def __init__(self, variable_names):
        self.variable_names = variable_names
        self.string_tuple = None
        self.value_cache = [None] * len(variable_names)

    def _reset(self, string_tuple):
        self.string_tuple = string_tuple
        value_cache = self.value_cache
        for i in range(len(self.variable_names)):
            value_cache[i] = None

    def _validate_index(self, index):
        if index >= 0 and index < len(self.string_tuple):
            return index

        raise IndexError("Out-of-bounds index passed to BindingSet." +
                         "  Index must be between 0 and %i, inclusive." % (len(self.string_tuple) - 1))

    def _get_ith_value(self, index):
        term = self.value_cache[index]
        if not term:
            def convert(x):
                if isinstance(x, list): return [convert(elt) for elt in x]
                else: return parse_term(x)
            term = convert(self.string_tuple[index])
            self.value_cache[index] = term
        return term

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._get_ith_value(self._validate_index(key))

        try:
            return self._get_ith_value(self.variable_names.index(key))
        except ValueError:
            raise KeyError(("Illegal key '%s' passed to binding set." +
                            "\n   Legal keys are %s") % (key, unicode(self.variable_names)))


    def iterator(self):
        """
        Creates an iterator over the bindings in this BindingSet. This only
        returns bindings with non-null values. An implementation is free to return
        the bindings in arbitrary order.

        Currently, we only support Python-style iteration over BindingSet dictionaries,
        so this (Java-style) iterator method is not implemented.
        """
        raise NotImplementedError("iterator")

    def getBindingNames(self):
        """
        Gets the names of the bindings in this BindingSet.

        """
        return self.variable_names

    def getBinding(self, bindingName):
        """
        Gets the binding with the specified name from this BindingSet.
        """
        return Binding(bindingName, self[bindingName])

    def hasBinding(self, bindingName):
        """
        Checks whether this BindingSet has a binding with the specified name.
        """
        return self.variable_names.find(bindingName) >= 0

    def getValue(self, bindingName):
        """
        Gets the value of the binding with the specified name from this BindingSet.
        Throws exception if 'bindingName' is not legal.
        """
        return self[bindingName]

    def getRow(self):
        """
        Return a list of strings representing the values of the current row.
        This exists mostly for debugging (otherwise, we would return terms instead of strings).
        """
        return self.string_tuple

    def __len__(self):
        return len(self.value_cache)

    def size(self):
        """
        Returns the number of bindings in this BindingSet.
        """
        return len(self)

    def _toDict(self, strings_dict=False):
        """
        Return a Python dictionary representation of this binding set.
        """
        d = {}
        for i in range(len(self.variable_names)):
            v = self[i]
            if strings_dict: v = unicode(v)
            d[self.variable_names[i]] = v
        return d

    def __str__(self):
        return unicode(self._toDict(strings_dict=True))


#############################################################################
##
#############################################################################

class Binding(namedtuple('Binding', 'name value')):
    __slots__ = ()
    """
    An implementation of 'Binding'
    """
    def getName(self):
        """
        Gets the name of the binding (e.g. the variable name).
        """
        return self.name

    def getValue(self):
        """
        Gets the value of the binding. The returned value is never equal to
        None, such a "binding" is considered to be unbound.
        """
        return self.value


#############################################################################
##



from franz.openrdf.rio.formats import Format


class TupleFormat(Format):
    """
    Serialization formats for tuples returned by queries.
    See :class:`.Format`.
    """
    # Map from extensions to formats, used by Format.format_for_file_name
    _ext_map = {}

    # These will be automatically converted to TupleFormat instances

    SPARQL = dict(
        name="SPARQL XML",
        mime_types=["application/sparql-results+xml", "application/xml"],
        file_extensions=["xml", "sx"])

    JSON = dict(
        name="SPARQL JSON",
        mime_types=["application/sparql-results+json", "application/json"],
        file_extensions=["json"])

    CSV = dict(
        name="SPARQL CSV",
        mime_types=["text/csv"],
        file_extensions=["csv"])

    TSV = dict(
        name="SPARQL TSV",
        mime_types=["text/tab-separated-values"],
        file_extensions=["tsv"])

    TABLE = dict(
        name="Table",
        mime_types=["text/table"],
        file_extensions=[])



################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import unicode_literals

from franz.openrdf.rio.formats import Format


class RDFFormat(Format):
    """
    Represents the concept of an RDF data serialization format. RDF formats are
    identified by a name and can have one or more associated
    MIME types, zero or more associated file extensions and can specify a
    default character encoding. Some formats are able to encode context
    information while others are not; this is indicated by the value of
    supports_contexts field. Similaraly, formats for which the
    supports_attributes flag is `True` are capable of encoding triple 
    attributes.
    """
    # A global dictionary mapping extensions to formats
    # Used by Format.format_for_file_name
    _ext_map = {}

    def __init__(self, name, mime_types=None, charset="UTF-8",
                 file_extensions=None, supports_namespaces=False,
                 supports_contexts=False, supports_attributes=False,
                 register=True):
        """
        Initialize a new RDF format object.

        :param name: Human-readable name of the format.
        :param mime_types: A list of MIME types used for this format.
                           The first element of this list will be used
                           as the content-type header during uploads.
        :param charset: Character set used by the format.
        :param file_extensions: List of file extensions for this format.
        :param supports_namespaces: If true, the format supports namespaces
                                    and qualified names. This has no impact
                                    on the Python API.
        :param supports_contexts: If True the format can store quads
                                  (and not just triples).
        :param supports_attributes: If True the format can represent triple
                                    attributes.
        :param register: If True file extensions will be added to the map
                         used by :meth:`Format.format_for_file_name`.
        """
        Format.__init__(self, name, mime_types, charset, file_extensions, register)
        self.supports_namespaces = supports_namespaces
        self.supports_contexts = supports_contexts
        self.supports_attributes = supports_attributes

    # These will be automatically converted to RDFFormat instances
        
    RDFXML = dict(
        name="RDF/XML",
        mime_types=["application/rdf+xml", "application/xml"],
        file_extensions=["rdf", "rdfs", "owl", "xml"],
        supports_namespaces=True, supports_contexts=False)

    NTRIPLES = dict(
        name="N-Triples",
        mime_types=["application/n-triples", "text/plain"],
        file_extensions=["nt", "ntriples"],
        supports_namespaces=False, supports_contexts=False)

    NQUADS = dict(
        name="N-Quads",
        mime_types=["application/n-quads"],
        file_extensions=["nq", "nquads"],
        supports_namespaces=False, supports_contexts=True)

    NQX = dict(
        name="Extended N-Quads (with attributes)",
        mime_types=["application/x-extended-nquads"],
        file_extensions=["nqx"],
        supports_namespaces=False, supports_contexts=True,
        supports_attributes=True)

    TURTLE = dict(
        name="Turtle",
        mime_types=["text/turtle"],
        file_extensions=["ttl", "turtle"],
        supports_namespaces=True,
        supports_contexts=False)

    TRIG = dict(
        name="TriG",
        mime_types=["application/trig"],
        file_extensions=["trig"],
        supports_namespaces=True,
        supports_contexts=True)

    TRIX = dict(
        name="TriX",
        mime_types=["application/trix"],
        file_extensions=["trix"],
        supports_namespaces=True,
        supports_contexts=True)

    TABLE = dict(
        name="Table",
        mime_types=["text/table"],
        file_extensions=[],
        supports_namespaces=False,
        supports_contexts=True)

    JSONLD = dict(
        name="JSON-LD",
        mime_types=["application/ld+json"],
        file_extensions=["json", "jsonld"],
        supports_namespaces=True,
        supports_contexts=True)



################################################################################
# Copyright (c) 2006-2017 Franz Inc.
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################
from __future__ import absolute_import, division, with_statement

import copy
import sys

if(sys.version_info.major < 3):
    from inspect import getargspec as getfullargspec
else:
    from inspect import getfullargspec

import math
import re
import threading
from contextlib import contextmanager
from datetime import timedelta

import six
from franz.miniclient.agjson import encode_json
from franz.openrdf.model.value import URI
from franz.openrdf.repository.attributes import AttributeFilter
from franz.openrdf.rio.formats import Format
from past.builtins import basestring
from past.utils import old_div
from six import python_2_unicode_compatible
from six.moves.urllib.parse import quote, urlparse

from ..openrdf.util.contexts import wrap_context
from ..openrdf.util.strings import to_native_string
from .request import (RequestError, decode, deserialize, encode, jsonRequest,
                      nullRequest, serialize, urlenc)


def _split_proxy(proxy):
    """
    Split a proxy definition of the form PROTOCOL://HOST:PORT into components.

    Return 3x None if the argument is None.

    :param proxy: A proxy specification string.
    :return: scheme, host and port as a triple.
    :rtype (string, string, int)
    """
    if proxy:
        url = urlparse(proxy)
        return url.scheme.lower(), url.hostname, url.port
    else:
        return None, None, None


class Service(object):
    def __init__(self, url, user=None, password=None, cainfo=None, sslcert=None,
                 verifyhost=None, verifypeer=None, proxy=None):
        # Note: it is important to save the arguments into fields with identical names.
        # This makes the weird cloning mechanism used by subclasses work.
        # WARNING: The cloning process mentioned above will create a **shallow** copy
        # of each argument.
        self.url = url
        self.user = user
        self.password = password
        self.runAsName = None
        self.cainfo = cainfo
        self.sslcert = sslcert
        self.verifyhost = verifyhost
        self.verifypeer = verifypeer
        self.proxy = proxy
        self.proxy_type, self.proxy_host, self.proxy_port = _split_proxy(proxy)
        self.session = None  # Will be created lazily
        self.transaction_settings = None
        self.user_attributes = None

    def getHeaders(self):
        """
        Return a dictionary of headers that should be added to every request.
        Return ``None`` instead of an empty dictionary if there are no such
        headers.
        """
        s = self.transaction_settings
        values = []
        if s:
            if s.distributed_transaction_timeout is not None:
                values.append(
                    ('distributedTransactionTimeout',
                     time_in_seconds(s.distributed_transaction_timeout)))
            if s.durability is not None:
                values.append(('durability', s.durability))
            if s.transaction_latency_count is not None:
                values.append(('transactionLatencyCount',
                               s.transaction_latency_count))
            if s.transaction_latency_timeout is not None:
                values.append(('transactionLatencyTimeout',
                               time_in_seconds(s.transaction_latency_timeout)))
        result = {}
        if values:
            result['x-repl-settings'] = ' '.join('%s=%s' % value for value in values)
        if self.user_attributes:
            result['x-user-attributes'] = encode_json(self.user_attributes)
        return result or None

    def _instance_from_url(self, subclass, url=None):
        """
        Create a copy of this service object that will:
           - be an instance of the ``subclass`` class.
           - share the values of all attributes except url,
             which is set to the specified value.

        This is used to create Repository and Catalog objects in the code below.
        """
        # Copy authentication and other settings
        kwargs = {}
        for arg_name in getfullargspec(subclass.__init__).args:
            if hasattr(self, arg_name):
                kwargs[arg_name] = copy.copy(getattr(self, arg_name))
        # Override URL
        if url is not None:
            kwargs['url'] = url
        return subclass(**kwargs)

    def copy(self):
        return self._instance_from_url(type(self))


class Catalog(Service):
    def listRepositories(self):
        """Returns the names of repositories in the catalog."""
        repos = jsonRequest(self, "GET", "/repositories")
        return [repo["id"] for repo in repos]

    def createRepository(self, name, indices=None):
        """Ask the server to create a new repository."""
        nullRequest(self, "PUT", "/repositories/" + quote(name) + "?" +
            urlenc(index=indices))
        return self.getRepository(name)

    def deleteRepository(self, name):
        """Delete a repository in this catalog."""
        nullRequest(self, "DELETE", "/repositories/" + quote(name))

    def getRepository(self, name):
        """Create an access object for a triple store."""
        return self._instance_from_url(Repository, to_native_string(self.url) + "/repositories/" + quote(name))


class Client(Service):
    def getVersion(self):
        return jsonRequest(self, "GET", "/version")

    def listCatalogs(self):
        return [cat["id"] if cat["id"] != "/" else None for cat in jsonRequest(self, "GET", "/catalogs")]

    def openCatalog(self, uriOrName):
        if re.match("^https?://", uriOrName):
            return self._instance_from_url(Catalog, uriOrName)
        else:
            return self.openCatalogByName(uriOrName)

    def openCatalogByName(self, name=None):
        url = to_native_string(self.url)
        if name:
            name = to_native_string(name)
            if name != "~":
                url += "/catalogs/" + quote(name)
        return self.openCatalog(url)

    def getInitfile(self):
        return jsonRequest(self, "GET", "/initfile")

    def setInitfile(self, content=None, restart=True):
        if (content is None):
            nullRequest(self, "DELETE", "/initfile")
        else:
            nullRequest(self, "PUT", "/initfile?" + urlenc(restart=restart), content)

    def listScripts(self):
        """
        List the registered scripts.
        """
        return jsonRequest(self, "GET", "/scripts")

    def addScript(self, module, code):
        nullRequest(self, "PUT", "/scripts/" + to_native_string(module),
            body=code)

    def getScript(self, module):
        return jsonRequest(self, "GET", "/scripts/" + to_native_string(module))

    def deleteScript(self, module):
        nullRequest(self, "DELETE", "/scripts/" + to_native_string(module))

    def openSession(self, spec, autocommit=False, lifetime=None, loadinitfile=False):
        url = jsonRequest(self, "POST", "/session?" +
                          urlenc(autoCommit=autocommit, lifetime=lifetime,
                                 loadInitFile=loadinitfile, store=spec))
        rep = self._instance_from_url(Repository, url)
        rep._enableSession(lifetime)
        return rep

    def listUsers(self):
        return jsonRequest(self, "GET", "/users")

    def addUser(self, name, password=None):
        assert password is not None or name == "anonymous"
        nullRequest(self, "PUT", "/users/" + quote(name) + "?" + urlenc(password=password))

    def deleteUser(self, name):
        nullRequest(self, "DELETE", "/users/" + quote(name))

    def changeUserPassword(self, name, password):
        nullRequest(self, "POST", "/users/" + quote(name) + "/password", password)

    def listUserAccess(self, name):
        return jsonRequest(self, "GET", "/users/" + quote(name) + "/access")

    def addUserAccess(self, name, read, write, catalog, repository):
        nullRequest(self, "PUT", "/users/" + quote(name) + "/access?" +
            urlenc(read=read, write=write, catalog=catalog, repository=repository))

    def deleteUserAccess(self, name, read, write, catalog, repository):
        nullRequest(self, "DELETE", "/users/" + quote(name) + "/access?" +
            urlenc(read=read, write=write, catalog=catalog, repository=repository))

    def listUserEffectiveAccess(self, name):
        return jsonRequest(self, "GET", "/users/" + quote(name) + "/effectiveAccess")

    def listUserPermissions(self, name):
        return jsonRequest(self, "GET", "/users/" + quote(name) + "/permissions")

    def listUserEffectivePermissions(self, name):
        return jsonRequest(self, "GET", "/users/" + quote(name) + "/effectivePermissions")

    def addUserPermission(self, name, _type):
        nullRequest(self, "PUT", "/users/" + quote(name) + "/permissions/" + _type)

    def deleteUserPermission(self, name, _type):
        nullRequest(self, "DELETE", "/users/" + quote(name) + "/permissions/" + _type)

    def listRoles(self):
        return jsonRequest(self, "GET", "/roles")

    def addRole(self, role):
        nullRequest(self, "PUT", "/roles/" + quote(role))

    def deleteRole(self, role):
        nullRequest(self, "DELETE", "/roles/" + quote(role))

    def listRolePermissions(self, role):
        return jsonRequest(self, "GET", "/roles/" + quote(role) + "/permissions")

    def addRolePermission(self, role, _type):
        nullRequest(self, "PUT", "/roles/" + quote(role) + "/permissions/" + _type)

    def deleteRolePermission(self, role, _type):
        nullRequest(self, "DELETE", "/roles/" + quote(role) + "/permissions/" + _type)

    def listRoleAccess(self, role):
        return jsonRequest(self, "GET", "/roles/" + quote(role) + "/access")

    def addRoleAccess(self, role, read, write, catalog, repository):
        nullRequest(self, "PUT", "/roles/" + quote(role) + "/access?" +
            urlenc(read=read, write=write, catalog=catalog, repository=repository))

    def deleteRoleAccess(self, role, read, write, catalog, repository):
        nullRequest(self, "DELETE", "/roles/" + quote(role) + "/access?" +
            urlenc(read=read, write=write, catalog=catalog, repository=repository))

    def listUserRoles(self, name):
        return jsonRequest(self, "GET", "/users/" + quote(name) + "/roles")

    def addUserRole(self, name, role):
        nullRequest(self, "PUT", "/users/" + quote(name) + "/roles/" + quote(role))

    def deleteUserRole(self, name, role):
        nullRequest(self, "DELETE", "/users/" + quote(name) + "/roles/" + quote(role))

    def listUserSecurityFilters(self, name, _type):
        return jsonRequest(self, "GET", "/users/" + quote(name) + '/security-filters/' + to_native_string(_type))

    def addUserSecurityFilter(self, name, _type, s=None, p=None, o=None, g=None):
        nullRequest(self, "POST", "/users/" + quote(name) + "/security-filters/" + to_native_string(_type) + "?" +
            urlenc(s=s, p=p, o=o, g=g))

    def deleteUserSecurityFilter(self, name, _type, s=None, p=None, o=None, g=None):
        nullRequest(self, "DELETE", "/users/" + quote(name) + "/security-filters/" + to_native_string(_type) + "?" +
            urlenc(s=s, p=p, o=o, g=g))

    def listRoleSecurityFilters(self, role, _type):
        return jsonRequest(self, "GET", "/roles/" + quote(role) + '/security-filters/' + to_native_string(_type))

    def addRoleSecurityFilter(self, role, _type, s=None, p=None, o=None, g=None):
        nullRequest(self, "POST", "/roles/" + quote(role) + "/security-filters/" + to_native_string(_type) + "?" +
            urlenc(s=s, p=p, o=o, g=g))

    def deleteRoleSecurityFilter(self, role, _type, s=None, p=None, o=None, g=None):
        nullRequest(self, "DELETE", "/roles/" + quote(role) + "/security-filters/" + to_native_string(_type) + "?" +
            urlenc(s=s, p=p, o=o, g=g))

    def getUserData(self, key):
        try:
            return jsonRequest(self, "GET",
                               "/users/%s/data/%s" % (quote(self.user, safe=''),
                                                      quote(key, safe='')),
                               )
        except RequestError as e:
            if e.status == 404:
                return None
            raise

    def setUserData(self, key, data):
        nullRequest(self, "PUT", "/users/%s/data/%s" % (quote(self.user, safe=''),
                                                        quote(key, safe='')),
                    body=data)

    def deleteUserData(self, key):
        nullRequest(self, "DELETE", "/users/%s/data/%s" % (quote(self.user, safe=''),
                                                           quote(key, safe='')))

    def listUserData(self):
        entries = jsonRequest(self, "GET",
                              "/users/%s/data/" % quote(self.user, safe=''))
        return [entry['id'] for entry in entries]


class Repository(Service):
    sessionAlive = None

    def getSize(self, context=None):
        """Returns the amount of triples in the repository."""
        return jsonRequest(self, "GET", "/size", urlenc(context=context))

    def listContexts(self):
        """Lists the contexts (named graphs) that are present in this repository."""
        return [t["contextID"] for t in jsonRequest(self, "GET", "/contexts")]

    def evalSparqlQuery(self, query, infer=False, context=None, namedContext=None, callback=None,
                        bindings=None, planner=None, checkVariables=None, count=False, accept=None, analyze=False,
                        analysisTechnique=None, analysisTimeout=None, update=False):
        """Execute a SPARQL query. Context can be None or a list of
        contexts -- strings in "http://foo.com" form or "null" for the
        default context. Return type depends on the query type. ASK
        gives a boolean, SELECT a {names, values} object containing
        lists of lists of terms. CONSTRUCT and DESCRIBE return a list
        of lists representing statements. Callback WILL NOT work on
        ASK queries."""
        if accept is None:
            accept="text/integer" if count else "application/json"
        if analyze:
            accept="text/plain"
        if bindings is not None:
            bindings = "".join(["&$" + quote(a) + "=" + quote(b.encode("utf-8")) for a, b in list(bindings.items())])
        return jsonRequest(self, "POST" if update else "GET", self.url,
                   urlenc(query=query, infer=infer, context=context, namedContext=namedContext,
                       planner=planner, checkVariables=checkVariables,
                       analyzeIndicesUsed=analyze, queryAnalysisTechnique=analysisTechnique,
                       queryAnalysisTimeout=analysisTimeout,
                       returnQueryMetadata=True) + (bindings or ""),
                       callback=callback,
                       accept=accept)

    def evalPrologQuery(self, query, infer=False, callback=None, limit=None, count=False, accept=None):
        """Execute a Prolog query. Returns a {names, values} object."""
        if accept is None:
            accept="text/integer" if count else "application/json"
        return jsonRequest(self, "POST", self.url,
                           urlenc(query=query, infer=infer, queryLn="prolog", limit=limit),
                           callback=callback,
                           accept=accept)

    def evalGraphqlQuery(self, query, default_prefix, infer, namespaces, variables, aliases):
        """Execute a GraphQL query. Returns a JSON object."""
        api = "/graphql"
        api_args = "&".join([ name + "=" + arg for name, arg in [
            ("default-prefix", default_prefix),
            ("infer", "false" if infer == False else infer),
            ("namespaces", namespaces), ("variables", variables),
            ("aliases", aliases)]
            if arg
        ])
        if api_args:
            api += "?" + api_args
        return jsonRequest(self, method="POST", url=self.url+api, body=query)

    def definePrologFunctors(self, definitions):
        """Add Prolog functors to the environment. Takes a string
        containing Lisp-syntax functor definitions (using the <-- and
        <- operators)."""
        nullRequest(self, "POST", "/functor", definitions)

    def evalJavaScript(self, code):
        """Evaluate JavaScript code in the server."""
        return jsonRequest(self, "POST", "/eval", code, 'text/javascript')

    def evalInServer(self, code):
        """Evaluate Common Lisp code in the server."""
        return jsonRequest(self, "POST", "/eval", code)

    def runAsUser(self, username=None):
        """
        Only for use as a superuser during a session.

        Runs requests on this connection as username.

        None - the default - clears the setting.
        """
        assert self.sessionAlive, "runAsUser can only be used on a session."
        self.runAsName = username

    def commit(self):
        nullRequest(self, "POST", "/commit")

    def rollback(self):
        nullRequest(self, "POST", "/rollback")

    def getStatements(self, subj=None, pred=None, obj=None, context=None, infer=False, callback=None,
                      limit=None, offset=None, accept=None, tripleIDs=False, count=False):
        """Retrieve all statements matching the given constraints.
        Context can be None or a list of contexts, as in
        evalSparqlQuery."""
        if subj == [] or pred == [] or obj == [] or context == []: return []
        subjEnd, predEnd, objEnd = None, None, None
        if isinstance(subj, tuple): subj, subjEnd = subj
        if isinstance(pred, tuple): pred, predEnd = pred
        if isinstance(obj, tuple): obj, objEnd = obj

        if accept is None:
            accept = "application/json"
            if count: accept = "text/integer"
            elif tripleIDs: accept = "application/x-quints+json"
        return jsonRequest(self, "GET", "/statements",
            urlenc(subj=subj, subjEnd=subjEnd, pred=pred, predEnd=predEnd,
                obj=obj, objEnd=objEnd, context=context, infer=infer,
                limit=limit, offset=offset),
            callback=callback, accept=accept)

    def getStatementsById(self, ids, returnIDs=True, accept=None, callback=None):
        if accept is None and returnIDs:
            accept = "application/x-quints+json" if returnIDs else "application/json"
        return jsonRequest(self, "GET", "/statements/id", urlenc(id=ids),
                           accept=accept, callback=callback)

    def addStatement(self, subj, pred, obj, context=None, attributes=None):
        """Add a single statement to the repository."""
        args = {
            'subj': subj,
            'pred': pred,
            'obj': obj,
            'context': context,
            'attributes': attributes and encode_json(attributes)
        }
        nullRequest(self, "PUT", "/statement?" + urlenc(**args))

    def deleteMatchingStatements(self, subj=None, pred=None, obj=None, context=None):
        """Delete all statements matching the constraints from the
        repository. Context can be None or a single graph name."""
        nullRequest(self, "DELETE", "/statements",
                    urlenc(subj=subj, pred=pred, obj=obj, context=context))

    def addStatements(self, quads, commitEvery=None):
        """Add a collection of statements to the repository. Quads
        should be an array of or five four-element arrays, where the fourth
        element, the graph name, may be None. The fifth element, if present,
        must be a string with a JSON-encoded dictionary of attribute values.
        """
        nullRequest(self, "POST", "/statements?" +
                    urlenc(commit=commitEvery),
                    encode_json(quads), content_type="application/json")

    def loadData(self, data, rdf_format, base_uri=None, context=None,
                 commit_every=None, content_encoding=None, attributes=None,
                 json_ld_store_source=None,
                 json_ld_context=None, allow_external_references=None,
                 external_reference_timeout=None):
        nullRequest(self, "POST",
                    "/statements?" + urlenc(
                        context=context,
                        baseURI=base_uri,
                        commit=commit_every,
                        attributes=attributes and encode_json(attributes),
                        jsonLdStoreSource=json_ld_store_source,
                        jsonLdContext=json_ld_context and fix_json_ld_context(json_ld_context),
                        externalReferences=allow_external_references,
                        extermalReferenceTimeout=external_reference_timeout),
                    data,
                    content_type=Format.mime_type_for_format(rdf_format),
                    content_encoding=content_encoding,)

    def loadFile(self, file, rdf_format, baseURI=None, context=None, serverSide=False,
                 commitEvery=None, content_encoding=None, attributes=None,
                 json_ld_store_source=None,
                 json_ld_context=None, allow_external_references=None,
                 external_reference_timeout=None):
        mime = Format.mime_type_for_format(rdf_format)

        if serverSide:
            # file is a server-side path, body should be empty
            body_context = wrap_context(None)
        elif isinstance(file, basestring) and not serverSide:
            # file is a local path, read it and send as body
            body_context = open(file, 'rb')
            file = None
        else:
            # We were passed a file-like object that we must not close,
            # so we'll just pass it as body.
            body_context = wrap_context(file)
            file = None

        with body_context as body:
            params = urlenc(file=file, context=context, baseURI=baseURI, commit=commitEvery,
                            attributes=attributes and encode_json(attributes),
                            jsonLdStoreSource=json_ld_store_source,
                            jsonLdContext=json_ld_context and fix_json_ld_context(json_ld_context),
                            externalReferences=allow_external_references,
                            extermalReferenceTimeout=external_reference_timeout)
            nullRequest(self, "POST", "/statements?" + params, body,
                        content_type=mime, content_encoding=content_encoding)

    def getBlankNodes(self, amount=1):
        return jsonRequest(self, "POST", "/blankNodes", urlenc(amount=amount))

    def deleteStatements(self, quads):
        """Delete a collection of statements from the repository."""
        nullRequest(self, "POST", "/statements/delete", encode_json(quads), content_type="application/json")

    def deleteStatementsById(self, ids):
        nullRequest(self, "POST", "/statements/delete?ids=true", encode_json(ids), content_type="application/json")

    def evalFreeTextSearch(self, pattern, index=None, infer=False, limit=None, offset=None):
        """Use free-text indices to search for the given pattern.
        Returns an array of statements."""
        return jsonRequest(self, "GET", "/freetext",
                           urlenc(pattern=pattern, infer=infer, limit=limit, offset=offset, index=index))

    def listFreeTextIndices(self):
        """List the names of free-text indices defined in this
        repository."""
        return jsonRequest(self, "GET", "/freetext/indices")

    def getFreeTextIndexConfiguration(self, index):
        """Returns a dictionary with fields \"predicates\",
        \"indexLiterals\", \"indexResources\", \"indexFields\",
        \"minimumWordSize\", \"stopWords\", and \"wordFilters\"."""
        return jsonRequest(self, "GET", "/freetext/indices/" + quote(index))

    def createFreeTextIndex(self, index, predicates=None, indexLiterals=None, indexResources=None,
                            indexFields=None, minimumWordSize=None, stopWords=None, wordFilters=None,
                            innerChars=None, borderChars=None, tokenizer=None):
        """Create a free-text index. predicates, if given, should be a
        list of resources. indexLiterals can be True, False, or a list
        of strings, indicating the literal types or langs to index.
        indexResources can be True, False, or \"short\". indexFields
        can be a list containing any combination of the elements
        \"subject\", \"predicate\", \"object\", and \"graph\".
        minimumWordSize is an integer, and stopWords a list of
        strings. wordFilter must be a list of filter names.

        "innerChars" and "borderChars" can be lists. "tokenizer" is
        a string. See documentation:

        http://www.franz.com/agraph/support/documentation/current/http-protocol.html#put-freetext-index"""
        # Not passing these at all causes the defaults to be used. So
        # when they are given, they should be passed with an empty
        # value.
        if stopWords == []: stopWords = ""
        if indexFields == []: indexFields = ""
        nullRequest(self, "PUT", "/freetext/indices/" + quote(index),
                    urlenc(predicate=predicates, indexLiterals=indexLiterals and True,
                           indexLiteralType=indexLiterals if isinstance(indexLiterals, list) else None,
                           indexResources=indexResources, indexField=indexFields,
                           minimumWordSize=minimumWordSize, stopWord=stopWords, wordFilter=wordFilters,
                           innerChars=innerChars, borderChars=borderChars,
                           tokenizer=tokenizer))

    def modifyFreeTextIndex(self, index, predicates=None, indexLiterals=None, indexResources=None,
                            indexFields=None, minimumWordSize=None, stopWords=None, wordFilters=None,
                            reIndex=None, innerChars=None, borderChars=None,
                            tokenizer=None):
        """Reconfigure a free-text index. Most arguments work as in
        createFreeTextIndex, except that here not passing them means
        'leave the old value'. reIndex controls whether old triples
        are re-indexed."""
        if stopWords == []: stopWords = ""
        if predicates == []: predicates = ""
        if indexFields == []: indexFields = ""
        if wordFilters == []: wordFilters = ""
        nullRequest(self, "POST", "/freetext/indices/" + quote(index),
                    urlenc(predicate=predicates, indexLiterals=indexLiterals and True,
                           indexLiteralType=indexLiterals if isinstance(indexLiterals, list) else None,
                           indexResources=indexResources, indexField=indexFields,
                           minimumWordSize=minimumWordSize, stopWord=stopWords, wordFilter=wordFilters,
                           reIndex=reIndex, innerChars=innerChars, borderChars=borderChars,
                           tokenizer=tokenizer))

    def deleteFreeTextIndex(self, index):
        """Delete the named free-text index."""
        nullRequest(self, "DELETE", "/freetext/indices/" + quote(index))

    def listFreeTextPredicates(self):
        """List the predicates that are used for free-text indexing."""
        return jsonRequest(self, "GET", "/freetext/predicates")

    def registerFreeTextPredicate(self, predicate):
        """Add a predicate for free-text indexing."""
        nullRequest(self, "POST", "/freetext/predicates", urlenc(predicate=predicate))

    def clearNamespaces(self, reset=True):
        """
        Deletes all namespaces in this repository for the current user. If a
        `reset` argument of `True` is passed, the user's namespaces are reset
        to the default set of namespaces, otherwise all namespaces are cleared.
        """
        nullRequest(self, "DELETE", "/namespaces?" + urlenc(reset=reset))

    def addNamespace(self, prefix, uri):
        nullRequest(self, "PUT", "/namespaces/" + quote(prefix),
                    uri, content_type="text/plain")

    def deleteNamespace(self, prefix):
        nullRequest(self, "DELETE", "/namespaces/" + quote(prefix))

    def listNamespaces(self):
        return jsonRequest(self, "GET", "/namespaces")

    def getNamespace(self, prefix):
        return jsonRequest(self, "GET", "/namespaces/" + quote(prefix))

    def clearQueryOptions(self):
        """
        Deletes all query options in this repository for the current user.
        """
        nullRequest(self, "DELETE", "/query-options")

    def setQueryOption(self, name, value):
        nullRequest(self, "PUT", "/query-options/" + quote(name),
                    value, content_type="text/plain")

    def deleteQueryOption(self, name):
        nullRequest(self, "DELETE", "/query-options/" + quote(name))

    def listQueryOptions(self):
        return jsonRequest(self, "GET", "/query-options")

    def getQueryOption(self, name):
        return jsonRequest(self, "GET", "/query-options/" + quote(name))

    def listMappedTypes(self):
        return jsonRequest(self, "GET", "/mapping/type")

    def addMappedType(self, type, encoding):
        nullRequest(self, "POST", "/mapping/type", urlenc(type=type, encoding=encoding))

    def deleteMappedType(self, type):
        nullRequest(self, "DELETE", "/mapping/type", urlenc(type=type))

    def listMappedPredicates(self):
        return jsonRequest(self, "GET", "/mapping/predicate")

    def addMappedPredicate(self, predicate, encoding):
        nullRequest(self, "POST", "/mapping/predicate",
                    urlenc(predicate=predicate, encoding=encoding))

    def deleteMappedPredicate(self, predicate):
        nullRequest(self, "DELETE", "/mapping/predicate", urlenc(predicate=predicate))

    def listIndices(self):
        return jsonRequest(self, "GET", "/indices")

    def listValidIndices(self):
        return jsonRequest(self, "GET", "/indices?listValid=true")

    def addIndex(self, _type):
        nullRequest(self, "PUT", "/indices/" + quote(_type))

    def dropIndex(self, _type):
        nullRequest(self, "DELETE", "/indices/" + quote(_type))

    def optimizeIndices(self, level=None, wait=None):
        nullRequest(self, "POST", "/indices/optimize?" + urlenc(
            level=level,wait=wait))

    def getCartesianGeoType(self, stripWidth, xMin, xMax, yMin, yMax):
        """Retrieve a cartesian geo-spatial literal type."""
        return jsonRequest(self, "POST", "/geo/types/cartesian?" +
                           urlenc(stripWidth=stripWidth, xmin=xMin, ymin=yMin, xmax=xMax, ymax=yMax))


    def getSphericalGeoType(self, stripWidth, unit="degree", latMin=None, latMax=None, longMin=None, longMax=None):
        """Retrieve a spherical geo-spatial literal type."""
        return jsonRequest(self, "POST", "/geo/types/spherical?" +
                           urlenc(stripWidth=stripWidth, unit=unit, latmin=latMin, latmax=latMax,
                                  longmin=longMin, longmax=longMax))

    def listGeoTypes(self):
        """List the geo-spatial types registered in the store."""
        return jsonRequest(self, "GET", "/geo/types")

    def createCartesianGeoLiteral(self, type, x, y):
        """Create a geo-spatial literal of the given type."""
        return "\"%+g%+g\"^^%s" % (x, y, type)

    @python_2_unicode_compatible
    class UnsupportedUnitError(Exception):
        def __init__(self, unit): self.unit = unit
        def __str__(self): return "'%s' is not a known unit (use km, mile, degree, or radian)." % self.unit

    def unitDegreeFactor(self, unit):
        if unit == "degree": return 1.0
        elif unit == "radian": return 57.29577951308232
        elif unit == "km": return 0.008998159
        elif unit == "mile": return 0.014481134
        else: raise Repository.UnsupportedUnitError(unit)

    def createSphericalGeoLiteral(self, type, lat, longitude, unit="degree"):
        """Create a geo-spatial latitude/longitude literal of the
        given type. Unit can be 'km', 'mile', 'radian', or 'degree'."""
        def asISO6709(number, digits):
            sign = "+"
            if (number < 0):
                sign= "-"
                number = -number
            fl = math.floor(number)
            return sign + (("%%0%dd" % digits) % fl) + (".%07d" % ((number - fl) * 10000000))

        conv = self.unitDegreeFactor(unit)
        return "\"%s%s\"^^%s" % (asISO6709(lat * conv, 2), asISO6709(longitude * conv, 3), type)

    def getStatementsHaversine(self, type, predicate, lat, longitude, radius, unit="km", limit=None, offset=None,
                               accept=None, callback=None):
        """Get all the triples with a given predicate whose object
        lies within radius units from the given latitude/longitude."""
        return jsonRequest(self, "GET", "/geo/haversine",
                           urlenc(type=type, predicate=predicate, lat=lat, long=longitude, radius=radius, unit=unit,
                                  limit=limit, offset=offset),
                           accept=accept, callback=callback)

    def getStatementsInsideBox(self, type, predicate, xMin, xMax, yMin, yMax, limit=None, offset=None,
                               accept=None, callback=None):
        """Get all the triples with a given predicate whose object
        lies within the specified box."""
        return jsonRequest(self, "GET", "/geo/box",
                           urlenc(type=type, predicate=predicate, xmin=xMin, xmax=xMax, ymin=yMin, ymax=yMax,
                                  limit=limit, offset=offset),
                           accept=accept, callback=callback)

    def getStatementsInsideCircle(self, type, predicate, x, y, radius, limit=None, offset=None,
                                  accept=None, callback=None):
        """Get all the triples with a given predicate whose object
        lies within the specified circle."""
        return jsonRequest(self, "GET", "/geo/circle",
                           urlenc(type=type, predicate=predicate, x=x, y=y, radius=radius, limit=limit, offset=offset),
                           accept=accept, callback=callback)

    def getStatementsInsidePolygon(self, type, predicate, polygon, limit=None, offset=None, accept=None, callback=None):
        """Get all the triples with a given predicate whose object
        lies within the specified polygon (see createPolygon)."""
        return jsonRequest(self, "GET", "/geo/polygon",
                           urlenc(type=type, predicate=predicate, polygon=polygon, limit=limit, offset=offset),
                           accept=accept, callback=callback)

    def createPolygon(self, resource, points):
        """Create a polygon with the given name in the store. points
        should be a list of literals created with createCartesianGeoLiteral."""
        nullRequest(self, "PUT", "/geo/polygon?" + urlenc(resource=resource, point=points))

    def registerSNAGenerator(self, name, subjectOf=None, objectOf=None, undirected=None, query=None):
        """subjectOf, objectOf, and undirected can be either a single
        predicate or a list of predicates. query should be a prolog
        query in the form (select ?x (q- ?node !<mypredicate> ?x)),
        where ?node always returns to the argument passed to the
        generator."""
        nullRequest(self, "PUT", "/snaGenerators/" + quote(name) + "?" +
                    urlenc(subjectOf=subjectOf, objectOf=objectOf, undirected=undirected, query=query))

    def registerNeighborMatrix(self, name, group, generator, depth):
        """group is a list of nodes, generator the name of an SNA generator."""
        nullRequest(self, "PUT", "/neighborMatrices/" + quote(name) + "?" +
                    urlenc(group=group, depth=depth, generator=generator))

    def getTripleCacheSize(self):
        return jsonRequest(self, "GET", "/tripleCache") or False

    def disableTripleCache(self):
        nullRequest(self, "DELETE", "/tripleCache")

    def enableTripleCache(self, size=None):
        nullRequest(self, "PUT", "/tripleCache?" + urlenc(size=size))

    def warmup(self, includeStrings, includeTriples, indices):
        nullRequest(self, "PUT", "/warmup?" + urlenc(includeStrings=includeStrings,
                                                     includeTriples=includeTriples,
                                                     index=indices))

    def openSession(self, autocommit=False, lifetime=None, loadinitfile=False):
        if self.sessionAlive:
            # A session is already active.  Do nothing.
            return
        self.oldUrl = self.url
        self.url = jsonRequest(self, "POST", "/session?" + urlenc(autoCommit=autocommit,
            lifetime=lifetime, loadInitFile=loadinitfile))
        self._enableSession(lifetime)

    def _enableSession(self, lifetime):
        alive = self.sessionAlive = threading.Event()

        def pingSession():
            while True:
                # Block until closeSession is called, or until
                # the polling interval has been reached.
                # NOTE: wait timeout is specified in seconds.
                alive.wait(old_div(lifetime, 2) if lifetime else 250)
                if alive.isSet():
                    # closeSession has been called.  Terminate the loop
                    # (and therefore the thread)
                    return
                try: nullRequest(self, "GET", "/session/ping")
                except Exception:
                    # The ping failed.  Terminate the loop
                    # (and therefore the thread)
                    return

        pinger = threading.Thread(target=pingSession)
        # Mark the pingSession thread as a daemon thread.  This will
        # prevent it from keeping Python alive on shutdown.  This flag
        # must be set before the thread is started.
        pinger.daemon = True
        pinger.start()

    def closeSession(self):
        if not self.sessionAlive: return
        # Notify pingSession() of session close
        self.sessionAlive.set()
        self.sessionAlive = None
        try: nullRequest(self, "POST", "/session/close")
        except Exception: pass
        if hasattr(self, "oldUrl"): self.url = self.oldUrl

    def setAutoCommit(self, on):
        """Only allowed when a session is active."""
        assert self.sessionAlive, "AutoCommit can only be set on a session."
        nullRequest(self, "POST", "/session/autoCommit?" + urlenc(on=on))

    def setBulkMode(self, on):
        nullRequest(self, "PUT" if on else "DELETE", "/bulkMode")

    def getBulkMode(self):
        return jsonRequest(self, "GET", "/bulkMode")

    def __del__(self):
        self.closeSession()

    def registerEncodedIdPrefix(self, prefix, uri_format):
        return nullRequest(self, "POST", "/encodedIds/prefixes?" +
                           urlenc(prefix=prefix, format=uri_format))

    def registerEncodedIdPrefixes(self, registrations):
        body = []
        for reg in registrations:
            body.append({
                "prefix":
                    reg.prefix if hasattr(reg, "prefix") else reg[0],
                "format":
                    reg.format if hasattr(reg, "format") else reg[1]})

        nullRequest(self, "POST", "/encodedIds/prefixes",
                    content_type="application/json", body=encode_json(body))

    def listEncodedIdPrefixes(self):
        return jsonRequest(self, "GET", "/encodedIds/prefixes")

    def unregisterEncodedIdPrefix(self, prefix):
        nullRequest(self, "DELETE", "/encodedIds/prefixes?" + urlenc(prefix=prefix))

    def allocateEncodedIds(self, prefix, amount):
        return jsonRequest(self, "POST", "/encodedIds?" + urlenc(prefix=prefix,
            amount=amount))

    def deleteDuplicates(self, mode):
        nullRequest(self, "DELETE", "/statements/duplicates?" + urlenc(mode=mode))

    def getDuplicateStatements(self, mode):
        accept = "application/json"
        return jsonRequest(self, "GET", "/statements/duplicates?" + urlenc(mode=mode), accept=accept)

    def getDuplicateSuppressionPolicy(self):
        return jsonRequest(self, "GET", "/suppressDuplicates", accept="application/json") or None

    def setDuplicateSuppressionPolicy(self, mode):
        nullRequest(self, "PUT", "/suppressDuplicates?" + urlenc(type=mode or "false"))

    def disableDuplicateSuppression(self):
        nullRequest(self, "DELETE", "/suppressDuplicates")

    def callStoredProc(self, function, module, *args):
        encoded = encode(serialize(args))
        return deserialize(decode(jsonRequest(self, "POST", "/custom/"+function,
            body=urlenc(spargstr=encoded), accept="text/plain",
            headers=["x-scripts: " + module])))

    def getSpinFunction(self, uri):
        """
        Gets the string of the function for the given uri.
         uri - Spin function identifier
        """
        return jsonRequest(self, "GET", "/spin/function/" + quote(uri, ''))

    def putSpinFunction(self, uri, sparqlQuery, arguments):
        """
        Adds a Spin function.
         uri - Spin function identifier
         sparqlQuery - Spin function query text
         arguments - names of arguments in the sparqlQuery
        """
        nullRequest(self, "PUT", "/spin/function/" + quote(uri, '') + "?" + urlenc(query=sparqlQuery, arguments=arguments))

    def deleteSpinFunction(self, uri):
        """
        Deletes the Spin function at the given uri.
         uri - Spin function identifier
        """
        nullRequest(self, "DELETE", "/spin/function/" + quote(uri, ''))

    def listSpinFunctions(self):
        """
        Returns a list of defined SPIN function.
        """
        return jsonRequest(self, "GET", "/spin/function")

    def putSpinMagicProperty(self, uri, sparqlQuery, arguments):
        """
        Add a Spin magic property.
         uri - Spin magic property identifier
         sparqlQuery
         arguments - names of arguments to the sparqlQuery - must contain the leading question mark
        """
        nullRequest(self, "PUT", "/spin/magicproperty/" + quote(uri, '') + "?" + urlenc(query=sparqlQuery, arguments=arguments))

    def getSpinMagicProperty(self, uri):
        """
        Get the spin magic property for the uri
         uri - spin magic property identifier
        """
        return jsonRequest(self, "GET", "/spin/magicproperty/" + quote(uri, ''))

    def listSpinMagicProperties(self):
        """
        Returns a list of defined SPIN magic properties function.
        """
        return jsonRequest(self, "GET", "/spin/magicproperty")

    def deleteSpinMagicProperty(self, uri):
        """
        Deletes the Spin magic property at the given uri.
         uri - Spin magic property identifier
        """
        nullRequest(self, "DELETE", "/spin/magicproperty/" + quote(uri, ''))

    def materializeEntailed(self, _with=None, without=None, useTypeSubproperty=False, commit=100000):
        """
        Call to materialize entailed triples to enable reasoning queries without the dynamic query-time reasoner.
        Returns the number of triples added.

        _with and without can be either a single string or a list of strings denoting rules beyond rdfs++ you wish to use.
        See the documentation for the current set, but "same-as", "restriction", "values-from", "class", and "property" are examples.

        useTypeSubproperty tells the materializer to prefer using types that are rdfs:subPropertyOf rdf:type rather than rdf:type directly.

        commit indicates the number of triples per commit for the materializer.
        """
        args = { "with": _with, "without": without, "useTypeSubproperty": useTypeSubproperty, "commit": commit }
        return jsonRequest(self, "PUT", "/materializeEntailed?" + urlenc(**args))

    def deleteMaterialized(self):
        """
        Deletes all previously materialized triples.
        Returns the number of triples deleted.
        """
        return jsonRequest(self, "DELETE", "/materializeEntailed")

    @contextmanager
    def saveResponse(self, fileobj, accept, raiseAll=False):
        """
        Save the server response(s) for the call(s) within the with statement
        to fileobj, using accept for the response type requested.
        """
        self._saveFile = fileobj
        self._saveAccept = accept
        try:
            yield
        except RequestError:
            raise
        except:
            if raiseAll:
                raise
        finally:
            del self._saveFile
            del self._saveAccept

    def setAttributeFilter(self, attribute_filter):
        if isinstance(attribute_filter, AttributeFilter):
            attribute_filter = attribute_filter.to_expr()
        args = urlenc(filter=attribute_filter)
        return nullRequest(self, 'POST', '/attributes/staticFilter?' + args)

    def getAttributeFilter(self):
        return jsonRequest(self, 'GET', '/attributes/staticFilter')

    def clearAttributeFilter(self):
        return nullRequest(self, 'DELETE', '/attributes/staticFilter')

    def getAttributeDefinitions(self):
        return jsonRequest(self, 'GET', '/attributes/definitions')

    def getAttributeDefinition(self, name):
        return jsonRequest(
            self, 'GET', '/attributes/definitions?' + urlenc(name=name))

    def setAttributeDefinition(self, attr_def):
        args = {
            'name': attr_def.name,
            'allowed-values': attr_def.allowed_values,
            'ordered': attr_def.ordered,
            'minimum-number': attr_def.minimum_number,
            'maximum-number': attr_def.maximum_number
        }
        return nullRequest(
            self, 'POST', '/attributes/definitions?' + urlenc(**args))

    def deleteAttributeDefinition(self, name):
        return jsonRequest(
            self, 'DELETE', '/attributes/definitions?' + urlenc(name=name))

    def loadDocument(self, doc, doc_format, base=None,
                     rules=None, subject=None, prefix=None,
                     rename=None, rdf_type=None, lang=None, skip=None,
                     transform=None, graph=None, json_store_source=None,
                     csv_columns=None, csv_separator=None, csv_quote=None,
                     csv_whitespace=None, csv_double_quote=None, csv_escape=None,
                     attributes=None, commit=None, context=None,
                     encoding=None, content_encoding=None):
        mime = Format.mime_type_for_format(doc_format)
        if encoding:
            mime += ';charset=' + encoding

        if prefix is None:
            prefix = {}
        if rename is not None:
            # Copy the keys, since we will be changing some renames
            for k, v in list(six.iteritems(rename)):
                if isinstance(v, URI):
                    # To provide a full URI for a key we need to use both 'prefix'
                    # and 'rename' options
                    namespace, local_name = v.split()
                    if local_name != k:
                        rename[k] = local_name
                    prefix[k] = namespace

        prefix_list = uri_dict_to_string_list(prefix) or []
        if base is not None:
            prefix_list.append(uri_to_string(base))

        args = {
            'transform-rules': rules,
            'tr-id': uri_to_string(subject),
            'tr-prefix': prefix_list,
            'tr-rename': dict_to_string_list(rename),
            'tr-type': uri_dict_to_string_list(rdf_type),
            'tr-lang':  dict_to_string_list(lang),
            'tr-skip': skip,
            'tr-transform':  dict_to_string_list(transform),
            'tr-graph': ['%s=%s' % (k, URI(str(v)))
                         for k, v in six.iteritems(graph or {})],
            'json-store-source': bool(json_store_source),
            'csv-columns': ','.join(csv_columns) if csv_columns else None,
            'csv-separator': csv_separator,
            'csv-quote': csv_quote,
            'csv-whitespace': ','.join(csv_whitespace) if csv_whitespace is not None else None,
            'csv-double-quote-p': bool(csv_double_quote) if csv_double_quote is not None else None,
            'csv-escape': csv_escape,
            'context': str(URI(str(context))) if context else None,
            'commit': commit,
            'attributes': attributes and encode_json(attributes)
        }
        return nullRequest(self, 'POST', '/statements/transform?' + urlenc(**args),
                           doc, content_type=mime, content_encoding=content_encoding)

    def getGeneration(self):
        return jsonRequest(self, 'GET', '/generation')


def uri_to_string(uri):
    """
    Converts a URI to a string. If the input is None then None is returned.
    If the input is a string it is returned after stripping angle brackets
    (if present).
    :param uri: Either a URI object or a string.
    :type uri: str|URI|None
    :return: str|None
    """
    if uri is None:
        return None
    elif isinstance(uri, URI):
        return uri.uri
    else:
        return URI(uri).uri


def uri_dict_to_string_list(d):
    """
    Convert a dict to a list of strings of the form key=value,
    but make sure that URI objects are serialized without angle brackets.
    """
    if d is None:
        return None
    result = []
    for k, v in six.iteritems(d):
        result.append('%s=%s' % (uri_to_string(k), uri_to_string(v)))
    return result


def dict_to_string_list(d):
    """
    Converts a dict to a list of 'key=value' strings. If the input is None
    then None will be returned.
    """
    return ['%s=%s' % kv for kv in six.iteritems(d)] if d else None


def time_in_seconds(t):
    """
    Try to convert an argument to a number of seconds.

    This works on:

        - integers (which are simply returned)
        - strings (cast to ints, suffixes like s, h, .. are *not* supported)
        - timedelta objects

    :param t: Time in seconds.
    :type t: int|str|timedelta
    :return: Number of seconds.
    :rtype: int
    """
    if t is None:
        return None
    if isinstance(t, timedelta):
        # Can't use total_seconds() in Python 2.6
        return t.seconds + t.days * 24 * 3600
    return int(t)


def fix_json_ld_context(context):
    """
    Take a dict or a string and return a JSON string.

    If the input dict does not contain "@context", wrap it.
    """
    if not isinstance(context, dict) or "@context" not in context:
        context = {"@context": context}
    return encode_json(context)



################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################
from datetime import timedelta, tzinfo


class MockTimezone(tzinfo):
    """
    A tzinfo implementation suitable for simple tests.
    """

    def __init__(self, name, utcoffset=timedelta(), dst=timedelta()):
        """
        Initialize a time zone.

        :param name: Display name.
        :type name: str
        :param utcoffset: Offset from the universal time.
                          Can be a timedelta or an int (number of hours).
                          Pass ``None`` to use the resulting tz to mark
                          'naive' time or datetime objects.
        :type utcoffset: timedelta|int|null
        :param dst: Time offset to account for DST.
                    This must be already included in ``utcoffset``.
                    In this test implementation it is constant and does not
                    depend on the date.
                    Can be a timedelta or an int (number of hours).
        :type dst: timedelta|int
        """
        tzinfo.__init__(self)
        self._name = name
        self._utcoffset = to_delta(utcoffset)
        self._dst = to_delta(dst)

    def tzname(self, date_time):
        return self._name

    def utcoffset(self, date_time):
        return self._utcoffset

    def dst(self, date_time):
        return self._dst


def to_delta(value):
    """
    Take either a timedelta or an int (number of hours),
    return a timedelta object.

    If the argument is ``None``, return ``None``.

    :type value: int|timedelta|null
    :rtype: timedelta
    """
    if value is None:
        return None
    elif isinstance(value, timedelta):
        return value
    return timedelta(hours=value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103 

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from past.builtins import basestring
from future.utils import python_2_unicode_compatible

from franz.openrdf.model.utils import parse_term
from .value import Value, URI


@python_2_unicode_compatible
class Statement(object):
    """
    Wraps a triple or a quad. Might also contain an id.
    """
    __slots__ = ('subject', 'predicate', 'object', 'context', 'id', '_hash')

    def __init__(self, subject, predicate, object, context=None, id=None):
        """
        Create a statement.

        Each component can be either a Value object or a string in N-Triples format.
        Strings will be parsed lazily the first time a component is accessed.
        Accessors will always return Value objects.

        :param subject: Subject component - URI, blank node or a string
                        in N-Triples format.
        :type subject: URI|BNode|str
        :param predicate: Predicate component. Either a URI or a string
                          in N-Triples format.
        :type predicate: URI|str
        :param object: Subject component. Either a Value or a string
                       in N-Triples format.
        :type object: Value|str
        :param context: Graph component (optional). Can be a URI or
                        a string in N-Triples format.
        :type context: URI|str
        :param id: Statement id (optional).
        :type id: int|str
        """
        self.subject = subject
        self.predicate = predicate
        self.object = object
        self.context = context
        self.id = id
        self._hash = None

    def __eq__(self, other):
        if not isinstance(other, Statement):
            return NotImplemented

        # The object is potentially the cheapest to check, as types
        # of these references might be different.
        # In general the number of different predicates in sets of
        # statements is the smallest, so predicate equality is checked
        # last.
        if self.getObject() == other.getObject() and self.getSubject() == other.getSubject() \
                and self.getPredicate() == other.getPredicate():
            if self.context:
                return self.getContext() == other.getContext()
            else:
                return not other.getContext()
        else:
            return False

    def __hash__(self):
        if self._hash is None:
            self._hash = hash((self.getSubject(), self.getPredicate(), self.getObject(), self.getContext()))
        return self._hash

    def __str__(self):
        elements = [self.getSubject(), self.getPredicate(), self.getObject(), self.getContext(), self.getTripleID()]
        while len(elements) > 3 and elements[-1] is None:
            elements.pop()
        return '(' + ', '.join(map(str, elements)) + ')'

    def __len__(self):
        return 3 if self.context is None else 4

    def __getitem__(self, index):
        if index == 0: return self.getSubject()
        elif index == 1: return self.getPredicate()
        elif index == 2: return self.getObject()
        elif index == 3: return self.getContext()
        else:
            raise IndexError('Illegal index (%d), must be < 4' % index)

    def getSubject(self):
        """
        Get the subject (the first element of the statement).

        :return: Subject.
        :rtype: Value
        """
        # Lazily parse and replace the value if needed.
        if isinstance(self.subject, basestring):
            self.subject = parse_term(self.subject)
        return self.subject
    
    def setSubject(self, subject):
        self.subject = subject
    
    def getPredicate(self):
        """
        Get the predicate (the second element of the statement).

        :return: Predicate.
        :rtype: URI
        """
        # Lazily parse and replace the value if needed.
        if isinstance(self.predicate, basestring):
            self.predicate = parse_term(self.predicate)
        return self.predicate
     
    def setPredicate(self, predicate):
        self.predicate = predicate
    
    def getObject(self):
        """
        Get the object (the third element of the statement).

        :return: Object.
        :rtype: Value
        """
        # Lazily parse and replace the value if needed.
        if isinstance(self.object, basestring):
            self.object = parse_term(self.object)
        return self.object
    
    def setObject(self, object):
        self.object = object
    
    def getContext(self):
        """
        Get the graph (the fourth, optional element of the statement).

        :return: Graph URI.
        :rtype: URI
        """
        # Lazily parse and replace the value if needed.
        if isinstance(self.context, basestring):
            self.context = parse_term(self.context)
        return self.context
    
    def setContext(self, context):
        self.context = context

    def getTripleID(self):
        """
        Get the statement id.

        Note that this field might not present, depending on the way in which
        the statement has been created.

        :return: A numerical id.
        :rtype: int
        """
        # Lazily parse and replace the value if needed.
        if isinstance(self.id, basestring):
            self.id = int(self.id)
        return self.id




# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

"""
Classes containing useful URI constants.
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from future.utils import iteritems

from franz.openrdf.model.value import URI
from .owl import OWL
from .rdf import RDF
from .rdfs import RDFS
from .xmlschema import XMLSchema

__all__ = ['OWL', 'RDF', 'RDFS', 'XMLSchema']



################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

"""
Helper function for opening connections.
"""
import os

from franz.openrdf.repository import Repository
from franz.openrdf.repository.repositoryconnection import RepositoryConnection
from franz.openrdf.sail import AllegroGraphServer


# Note that the default values come from AllegroGraphServer.__init__
def ag_connect(repo, catalog=None,
               create=True, fail_if_exists=False, clear=False,
               session=False, autocommit=False, lifetime=None, loadinitfile=False,
               host=None, port=None, protocol=None,
               user=None, password=None, cainfo=None, sslcert=None,
               verifyhost=None, verifypeer=None, indices=None,
               proxy=os.environ.get('AGRAPH_PROXY')):
    """
    Create a connection to an AllegroGraph repository.

    When closed the connection will take care of releasing all intermediate resources
    that were created in order to open it.

    :param repo: Repository name.
    :type repo: string
    :param catalog: Catalog name (optional, root catalog is the default).
    :type catalog: string
    :param create: if `True` (default) create the repository if it does not exist.
    :type create: bool
    :param fail_if_exists: if `True` and the repository exists raise an exception.
                           This applies only if `create` is `True`.
                           The default value is `False`.
    :type fail_if_exists: bool
    :param clear: if `True` delete all data after creating the connection.
                  The default is `False`.
    :type clear: bool
    :param session: If ``True`` start a session after creating the connection.
                    The default is ``False``.
    :type session: bool
    :param autocommit: When opening a session: if ``True``, commits are done on each
                       request, otherwise you will need to call :meth:`.commit` or
                       :meth:`.rollback` as appropriate for your application.
                       The default value is ``False``.
    :type autocommit: bool
    :param lifetime: Time (in seconds) before the session expires when idle.
                     Note that the client maintains a thread that ping the
                     session before this happens.
                     Ignored if not starting a session.
    :type lifetime: int
    :param loadinitfile: if ``True`` then the current initfile will be loaded
                         for you when the session starts. The default is ``False``.
                         Ignored if not starting a session.
    :type loadinitfile: bool
    :param host: AllegroGraph server host (default: ``'127.0.0.1'`` or the value
                 of the AGRAPH_HOST environment variable if that is defined.`
                 Can also be used to supply protocol and port number
                 (e.g. ``https://localhost:10036``).
    :type host: string
    :param protocol: Either ``"http"`` or ``"https"``.
                     The default is ``"http"``.
                     Overrides the protocol specified in ``host``.
    :type protocol: string
    :param port: AllegroGraph server port (default: `10035` for http and `10036`
                 for https, or the AGRAPH_PORT environment variable if that is defined).
                 Overrides the port number provided in ``host``.
    :type port: int
    :param user: Username for authentication (default: value of the ``AGRAPH_USER``
                 environment variable).
    :type user: string
    :param password:  Password for authentication (default: value of the ``AGRAPH_PASSWORD``
                      environment variable).
    :type password: string
    :param cainfo: Path to file or directory with CA certificates.
    :type cainfo: string
    :param sslcert: Path to a client certificate to use for
                    authentication instead of username and password.
    :type sslcert: string
    :param verifyhost: See https://curl.haxx.se/libcurl/c/CURLOPT_SSL_VERIFYHOST.html
    :type verifyhost: int
    :param verifypeer: See https://curl.haxx.se/libcurl/c/CURLOPT_SSL_VERIFYPEER.html
    :type verifypeer: int
    :param indices: List of indices to create if creating a new repository.
    :type indices: list[string]
    :param proxy: Proxy specification string. The format is SCHEME://HOST:PORT.
                  Supported schemes are 'http', 'socks4' and 'socks5'.
                  Note that for SOCKS proxies DNS requests are performed by the
                  proxy server.
                  The default value is taken from the AGRAPH_PROXY environment
                  variable.
    :type proxy: string
    :return: A :class:`.RepositoryConnection` object.
    :rtype: franz.openrdf.repositoryconnection.RepositoryConnection
    """
    server = AllegroGraphServer(host=host, port=port, protocol=protocol,
                                user=user, password=password, sslcert=sslcert,
                                cainfo=cainfo, verifyhost=verifyhost, verifypeer=verifypeer,
                                proxy=proxy)
    cat_handle = server.openCatalog(catalog)
    repo_exists = repo in cat_handle.listRepositories()
    if not repo_exists:
        if create:
            repo_handle = cat_handle.createRepository(repo, indices)
        else:
            raise Exception('Store %s does not exist.' % repo)
    else:
        if fail_if_exists and create:
            raise Exception('Store %s already exists.' % repo)
        mode = Repository.RENEW if clear else Repository.OPEN
        repo_handle = cat_handle.getRepository(repo, mode)
    conn = RepositoryConnection(repo_handle, close_repo=True)
    if session:
        # conn.close will close it if necessary
        conn.openSession(autocommit=autocommit, lifetime=lifetime, loadinitfile=loadinitfile)
    return conn



"""
Types and utilities related to transaction settings.
"""
from collections import namedtuple
try:
    from itertools import zip_longest  # Py3
except ImportError:
    from itertools import izip_longest as zip_longest  # Py2


# The subclass is created to allow us to provide a docstring for Sphinx.
class TransactionSettings(namedtuple('TransactionSettings', (
        'distributed_transaction_timeout', 'durability',
        'transaction_latency_count', 'transaction_latency_timeout'))):
    """
    A named tuple encapsulating a set of distributed transaction parameters.

    Each parameter here can be set to ``None``, in which case the actual value
    of that setting will be determined by the server.

    When creating a new instance not all arguments have to be specified,
    the default value for all settings is ``None``.

    Attributes:
        durability: (int|str):
            The durability is a positive integer value that specifies how many
            instances must have a commit ingested in order for that commit to be
            considered durable. The count includes the instance that made
            the commit.

            A durability setting of 1 means that when an instance makes a commit that
            commit is immediately considered durable before even being sent to any other
            instance (the commit will still be sent to the other instances after it's
            considered durable).

            A value that equals the total number of nodes in the cluster means that every
            instance must have ingested the commit before it's considered durable.
            If one or more instances are stopped at the moment then the commit will not
            become durable until the stopped instances are restarted.

            Three symbolic values (strings) may be used instead of numbers:
                - 'min', which simply means 1.
                - 'max', which means that all instances in the cluster must
                  ingest a commit before it is considered durable.
                - 'quorum', which means that more than a half of instances
                  must ingest a commit for it to be considered durable.

        distributed_transaction_timeout (int|datetime.timedelta):
            Use this setting to specify how long a commit call will wait for
            the commit to become durable. It's a non-negative integer
            (number of seconds).

            If the durability is greater than one then the committing process
            has to wait for acknowledgements from the other servers that the
            transaction was committed. The committing process returns to the
            caller when the durability has been reached or the distributed
            transaction timeout seconds has passed, whichever comes first.

            When the commit returns the caller does not know if durability has
            been reached.

        transaction_latency_count (int):
            Number of commits. Use this setting to limit the number of non-durable
            (pending) commits that can be active on the cluster. If this limit is
            reached all new commits will signal an error (and have to be retried).

            When a commit is done the committing process tries to wait until the
            commit is durable but if that takes too long (see
            distributed_transaction_timeout) then commit will return with the
            system still working on making that transaction durable.

            If the latency count is 4 then even if the last four commits are not yet
            durable it is possible to do one more commit. But if there are five
            pending transactions then any attempt to commit will result in an error.

            Another example: If you set the latency count to zero then each commit must
            be durable before the next commit can be ingested.

        transaction_latency_timeout (int|datetime.timedelta):
            Use this setting to specify how long a commit operation should wait for the
            transaction_Latency_count to be satisfied before throwing an error.

    """
    __slots__ = ()

    def __new__(cls, *args, **kwargs):
        """
        A constructor that provides default values.

        by default a named tuple requires all fields to be provided, this
        constructor will treat all missing values as ``None``.
        """
        params = dict(zip_longest(TransactionSettings._fields, args))
        params.update(kwargs)
        return super(TransactionSettings, cls).__new__(cls, **params)


DEFAULT_TRANSACTION_SETTINGS = TransactionSettings()


from __future__ import unicode_literals

import os


def assert_env_var_defined(name):
    value = os.environ.get(name)
    if value is None:
        raise Exception('{0} is not defined'.format(name))
    print('{name}: {value}'.format(name=name, value=value))


def setup():
    assert_env_var_defined('AGRAPH_PORT')
    assert_env_var_defined('AGRAPH_SSL_PORT')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from future.builtins import object
from past.builtins import unicode

from future.utils import iteritems

from ..model.value import URI

NS = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"

class RDF(object):
    """
    A 'static' class containing useful RDF URIs.
    """

    NAMESPACE = NS
    TYPE = URI(namespace=NS, localname="type")
    PROPERTY = URI(namespace=NS, localname="Property")
    XMLLITERAL = URI(namespace=NS, localname="XMLLiteral")
    SUBJECT = URI(namespace=NS, localname="subject")
    PREDICATE = URI(namespace=NS, localname="predicate")
    OBJECT = URI(namespace=NS, localname="object")
    STATEMENT = URI(namespace=NS, localname="Statement")
    BAG = URI(namespace=NS, localname="Bag")
    ALT = URI(namespace=NS, localname="Alt")
    SEQ = URI(namespace=NS, localname="Seq")
    VALUE = URI(namespace=NS, localname="value")
    LI = URI(namespace=NS, localname="li")
    LIST = URI(namespace=NS, localname="List")
    FIRST = URI(namespace=NS, localname="first")
    REST = URI(namespace=NS, localname="rest")
    NIL = URI(namespace=NS, localname="nil")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import unicode_literals
from franz.miniclient.request import RequestError

class IllegalOptionException(Exception):
    pass

class IllegalArgumentException (Exception):
    pass

class RDFHandlerException(Exception):
    pass

class ServerException(Exception):
    pass

class QueryMissingFeatureException(Exception):
    """
    Source language evokes a feature not supported by the execution language
    """



################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import print_function, unicode_literals

import sys

# JSON abstraction layer

try:
    import simplejson as json
except ImportError:
    import json


class JsonDecodeError(Exception):
    pass


def encode_json(value):
    return json.dumps(value)


def decode_json(text):
    # JSON on Py3 insists on getting Unicode strings
    if sys.version_info[0] > 2 and isinstance(text, bytes):
        text = text.decode('utf-8')
    try:
        return json.loads(text)
    except ValueError:
        raise JsonDecodeError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from past.builtins import unicode

from ..exceptions import IllegalArgumentException

## Finds the index of the first local name character in an (non-relative)
## URI. This index is determined by the following the following steps:
## <ul>
## <li>Find the <em>first</em> occurrence of the '#' character,
## <li>If this fails, find the <em>last</em> occurrence of the '/'
## character,
## <li>If this fails, find the <em>last</em> occurrence of the ':'
## character.
## <li>Add <tt>1<tt> to the found index and return this value.
## </ul>
## Note that the third step should never fail as every legal (non-relative)
## URI contains at least one ':' character to seperate the scheme from the
## rest of the URI. If this fails anyway, the method will throw an
## {@link IllegalArgumentException}.
## 
## @param uri
##        A URI string.
## @return The index of the first local name character in the URI string. Note that
## this index does not reference an actual character if the algorithm determines
## that there is not local name. In that case, the return index is equal to the
## length of the URI string.
## @throws IllegalArgumentException
##         If the supplied URI string doesn't contain any of the separator
##         characters. Every legal (non-relative) URI contains at least one
##         ':' character to separate the scheme from the rest of the URI.
def getLocalNameIndex(uri):
    idx = uri.rfind('#')
    if (idx < 0):
        idx = uri.rfind('/')
    if (idx < 0):
        idx = uri.rfind(':')
    if (idx < 0):
        raise IllegalArgumentException("No separator character found in URI: " + uri)
    return idx + 1

def asURIString(value):
    value = unicode(value)
    if value.startswith('<'): return value
    else: return "<%s>" % value



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103 

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals

import sys

import builtins

import weakref
from six import python_2_unicode_compatible
import six

from ..util import uris, strings


# Keys used to establish ordering between different types of terms
LITERAL_CMP_KEY = 1
URI_CMP_KEY = 2
BNODE_CMP_KEY = 3


@python_2_unicode_compatible
class Value(object):
    """
    Top class in the org.openrdf.model interfaces.
    """
    __slots__ = ('__weakref__',)

    def __str__(self):
        return self.toNTriples()

    def __repr__(self):
        result = self.toNTriples()
        # Return an ascii string on Python 2, otherwise pytest gets confused
        if sys.version_info < (3,):
            result = result.encode('unicode-escape')
        return result

    def get_cmp_key(self):
        """
        Return a key that will be used to compare and hash this object.
        """
        raise NotImplementedError()

    # Comparison methods rely on get_cmp_key.
    def __eq__(self, other):
        return self is other or (isinstance(other, Value)
                                 and self.get_cmp_key() == other.get_cmp_key())

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.get_cmp_key())

    def __lt__(self, other):
        return self.get_cmp_key() < other.get_cmp_key()

    def __gt__(self, other):
        return self.get_cmp_key() > other.get_cmp_key()

    def __le__(self, other):
        return not self > other

    def __ge__(self, other):
        return not self < other

    def toNTriples(self):
        """
        Return an NTriples representation of an open rdf term
        """
        raise NotImplementedError("Failed to implement 'toNTriples' on instance of type %s" % type(self).__name__)
    
    
class Resource(Value):
    __slots__ = ()
    pass


class URI(Resource):
    """
    Lightweight implementation of the class 'URI'.
    """
    __slots__ = ('_uri', '_is_canonical')

    _instances = weakref.WeakValueDictionary()

    def __new__(cls, uri=None, namespace=None, localname=None, canonical=True):
        if isinstance(uri, URI):
            if not canonical or uri._is_canonical:
                return uri
            uri = uri.uri
        elif uri is None and namespace is not None:
            uri = namespace + (localname or '')

        if uri is None:
            raise ValueError('Either URI or namespace is required.')

        if canonical:
            result = URI._instances.get(uri)
            if result is not None:
                return result

        result = super(URI, cls).__new__(cls)
        if canonical:
            URI._instances[uri] = result
        return result

    def __init__(self, uri=None, namespace=None, localname=None, canonical=False):
        if isinstance(uri, URI):
            uri = uri.uri

        if uri is None and namespace is not None:
            uri = namespace + (localname or '')

        if uri and uri[0] == '<' and uri[len(uri) - 1] == '>':
            # be kind and trim the uri:
            uri = uri[1:-1]
        self._uri = uri
        self._is_canonical = canonical

    def get_cmp_key(self):
        return URI_CMP_KEY, self.uri
    
    def getURI(self):
        """
        Return the URI (string) for 'self'.  This method is typically
        overloaded by subclasses, which may use lazy evaluation to
        retrieve the string.
        """
        return self._uri

    uri = property(getURI)
    
    def getValue(self):
        return self.getURI()

    value = property(getValue)
    
    def getLocalName(self):
        pos = uris.getLocalNameIndex(self.getURI())
        return self.uri[pos:]
    
    localname = property(getLocalName)

    def getNamespace(self):
        pos = uris.getLocalNameIndex(self.getURI())
        return self.uri[:pos]

    def split(self):
        """
        Split into a namespace + local name pair.
        """
        pos = uris.getLocalNameIndex(self.uri)
        return self.uri[:pos], self.uri[pos:]

    namespace = property(getNamespace)

    def toNTriples(self):
        """
        Return an NTriples representation of a resource, in this case, wrap
        it in angle brackets.
        """
        return strings.encode_ntriple_uri(self.uri)

    def to_json_ld_key(self):
        """ Converts to a string to be used as a JSON-LD key. """
        return self.uri

    def to_json_ld(self):
        """ Converts to an object to be used as a JSON-LD value. """
        return {"@id": self.uri}


class BNode(Resource):
    """
    A blank node.
    """
    __slots__ = ('id',)

    def __init__(self, id=None):
        """
        Create a blank node.

        :param id: Node identifier, if not supplied one will be generated.
        """
        if id is None:
            id = 'b' + hex(builtins.id(self))
        self.id = id
        
    def getId(self):
        """
        Get the identifier of this blank node.
        """
        return self.id

    getID = getId

    getValue = getID
    
    def get_cmp_key(self):
        return BNODE_CMP_KEY, self.getId()
    
    def toNTriples(self):
        return "_:%s" % self.getId()

    def to_json_ld_key(self):
        """ Converts to a string to be used as a JSON-LD key. """
        return self.toNTriples()

    def to_json_ld(self):
        """ Converts to an object to be used as a JSON-LD value. """
        return {"@id": self.toNTriples()}


@python_2_unicode_compatible
class Namespace(object):
    """
    """
    def __init__(self, prefix, name):
        self.setPrefix(prefix)
        self.setName(name)
    
    def getName(self):
        """
        Gets the name of the current namespace (i.e. it's URI).
        """
        return self.name

    def setName(self, name): self.name = name
    
    def getPrefix(self):
        """
        Gets the prefix of the current namespace.
        """
        return self.prefix

    def setPrefix(self, prefix): self.prefix = prefix
    
    def __str__(self):
        """
        Return an odd name (that's what the RDF4J code does).
        """
        return self.prefix + " :: " + self.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable-msg=C0103

################################################################################
# Copyright (c) 2006-2017 Franz Inc.  
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

from __future__ import absolute_import
from __future__ import unicode_literals
from builtins import object

from .rdfformat import RDFFormat

class RDFWriter(object):
    def __init__(self, rdfFormat, filePath=None):
        self.rdf_format = rdfFormat
        self.file_path = filePath
        
    def getRDFFormat(self):
        return self.rdf_format
    
    def getFilePath(self):
        return self.file_path 

            
class NTriplesWriter(RDFWriter):
    """
    Records the format as
    NTriples, and records the 'filePath' where the serialized RDF will
    be output to.  If 'filePath' is None, output is to standard output.
    
    TODO: THERE IS A WRITER PROTOCOL IMPLEMENTED IN RDFXMLWriter THAT ISN'T
    IMPLEMENTED HERE.  CONSIDER ADDING IT (OR NOT).
    """
    def __init__(self, filePath=None):
        super(NTriplesWriter, self).__init__(RDFFormat.NTRIPLES, filePath)
